"""Account Management plugin for VSS CLI (vss-cli)."""
import logging
from pathlib import Path
import sys
from typing import Tuple

import click

from vss_cli import autocompletion, const
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output
from vss_cli.utils.emoji import EMOJI_UNICODE
from vss_cli.validators import validate_phone_number

ej_tada = EMOJI_UNICODE.get(':party_popper:')
ej_mail = EMOJI_UNICODE.get(':closed_mailbox_with_raised_flag:')

_LOGGING = logging.getLogger(__name__)


def get_endpoint_and_creds(ctx: Configuration) -> Tuple[str, str, str]:
    """Get endpoint credentials."""
    endpoint = ctx.endpoint or click.prompt(
        'Endpoint',
        default=const.DEFAULT_ENDPOINT,
        type=click.STRING,
        show_default=True,
        err=True,
    )
    username = ctx.username or click.prompt(
        'Username',
        default=ctx.username,
        show_default=True,
        type=click.STRING,
        err=True,
    )
    password = ctx.password or click.prompt(
        'Password',
        default=ctx.password,
        show_default=False,
        hide_input=True,
        type=click.STRING,
        confirmation_prompt=True,
        err=True,
    )
    return endpoint, username, password


@click.group('account', short_help='Manage your VSS account')
@click.option(
    '--no-load', is_flag=True, default=False, help='do not load config'
)
@pass_context
def cli(ctx: Configuration, no_load: bool):
    """Manage your VSS account."""
    with ctx.spinner(disable=ctx.debug):
        if not no_load:
            ctx.load_config()


@cli.group('get', short_help='get account attribute')
@pass_context
def account_get(ctx: Configuration):
    """Obtain an account attribute."""
    pass


@account_get.group('digest')
@pass_context
def account_get_digest(ctx):
    """Get digest status."""
    pass


@account_get_digest.command('message')
@pass_context
def account_get_digest_message(ctx: Configuration):
    """Get message digest status."""
    with ctx.spinner(disable=ctx.debug):
        _obj = ctx.get_user_digest_settings()
    obj = {'message': _obj.get('message')}
    columns = ctx.columns or const.COLUMNS_MESSAGE_DIGEST
    ctx.echo(format_output(ctx, [obj], columns=columns, single=True))


@account_get.command('mfa')
@pass_context
def account_get_mfa(ctx: Configuration):
    """Get MFA account settings."""
    with ctx.spinner(disable=ctx.debug):
        _obj = ctx.get_user_totp()
    columns = ctx.columns or const.COLUMNS_USER_MFA
    ctx.echo(format_output(ctx, [_obj], columns=columns, single=True))


@account_get.command('groups')
@pass_context
def account_get_groups(ctx: Configuration):
    """User group membership."""
    with ctx.spinner(disable=ctx.debug):
        objs = ctx.get_user_groups(per_page=20)
    columns = ctx.columns or const.COLUMNS_GROUPS
    ctx.echo(format_output(ctx, objs, columns=columns))


@account_get.group('group', invoke_without_command=True)
@click.argument(
    'group_id_or_name',
    type=click.STRING,
    required=True,
    shell_complete=autocompletion.groups,
)
@pass_context
def account_get_group(ctx: Configuration, group_id_or_name):
    """Get given group info or members.

    User must be part of the group.
    """
    _group = ctx.get_vss_groups_by_name_desc_or_id(group_id_or_name)
    ctx.group = _group[0]['id']
    if click.get_current_context().invoked_subcommand is None:
        with ctx.spinner(disable=ctx.debug):
            obj = ctx.get_group(ctx.group)
        columns = ctx.columns or const.COLUMNS_GROUP
        ctx.echo(format_output(ctx, [obj], columns=columns, single=True))


@account_get_group.command('member')
@pass_context
def account_get_group_members(ctx: Configuration):
    """Get group members."""
    with ctx.spinner(disable=ctx.debug):
        objs = ctx.get_group_members(ctx.group)
    columns = ctx.columns or const.COLUMNS_GROUP_MEMBERS
    ctx.echo(format_output(ctx, objs, columns=columns))


@account_get.command('access-role')
@pass_context
def account_get_access_role(ctx: Configuration):
    """Access role and entitlements."""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_roles()
    columns = ctx.columns or const.COLUMNS_ROLE
    ctx.echo(format_output(ctx, [obj['access']], columns=columns, single=True))


@account_get.command('request-role')
@pass_context
def account_get_request_role(ctx: Configuration):
    """Request role and entitlements."""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_roles()
    columns = ctx.columns or const.COLUMNS_ROLE
    ctx.echo(
        format_output(ctx, [obj['request']], columns=columns, single=True)
    )


@account_get.command('personal')
@pass_context
def account_get_personal(ctx: Configuration):
    """Get user information."""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_personal()
    obj.update(ctx.get_user_ldap())
    columns = ctx.columns or const.COLUMNS_USER_PERSONAL
    ctx.echo(format_output(ctx, [obj], columns=columns, single=True))


@account_get.command('status')
@pass_context
def account_get_pstatus(ctx: Configuration):
    """Get account status."""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_status()
    columns = ctx.columns or const.COLUMNS_USER_STATUS
    ctx.echo(format_output(ctx, [obj], columns=columns, single=True))


@account_get.group('notification')
@pass_context
def account_get_notification(ctx: Configuration):
    """Get notification settings."""
    pass


@account_get_notification.command('request')
@pass_context
def account_get_notification_request(ctx: Configuration):
    """Get notification format."""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_request_notification_settings()
    columns = ctx.columns or const.COLUMNS_NOT_REQUEST
    ctx.echo(format_output(ctx, [obj], columns=columns, single=True))


@account_get_notification.command('format')
@pass_context
def account_get_notification_format(ctx: Configuration):
    """Get notification format."""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_notification_format()
    ctx.echo(format_output(ctx, [obj], columns=ctx.columns, single=True))


@account_get_notification.command('method')
@pass_context
def account_get_notification_method(ctx: Configuration):
    """Get notification format."""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_notification_method()
    ctx.echo(format_output(ctx, [obj], columns=ctx.columns, single=True))


@cli.group('set', short_help='set account attribute')
@pass_context
def account_set(ctx: Configuration):
    """Set account attribute."""
    pass


@account_set.group(
    'digest', short_help='set account weekly digest configuration'
)
@pass_context
def account_set_digest(ctx: Configuration):
    """Update weekly digest configuration."""
    pass


@account_set_digest.command('message')
@click.argument('state', type=click.Choice(['in', 'out']), required=True)
@pass_context
def account_set_digest_message(ctx: Configuration, state):
    """Opt-in or opt-out of weekly message digest."""
    with ctx.spinner(disable=ctx.debug):
        if state == 'in':
            ctx.enable_user_message_digest()
        else:
            ctx.disable_user_message_digest()
    _obj = ctx.get_user_digest_settings()
    obj = {'message': _obj.get('message')}
    columns = ctx.columns or const.COLUMNS_MESSAGE_DIGEST
    ctx.echo(format_output(ctx, [obj], columns=columns, single=True))


@account_set.group(
    'notification', short_help='set account notification settings'
)
@pass_context
def account_notification_set(ctx: Configuration):
    """Set account notification settings."""
    pass


@account_notification_set.command('request')
@click.argument(
    'notification_type',
    type=click.Choice(['all', 'none', 'error', 'completion', 'submission']),
    nargs=-1,
    required=True,
)
@pass_context
def account_notification_set_request(ctx: Configuration, notification_type):
    """Customize request notification settings."""
    lookup = {
        'all': ctx.enable_user_request_all_notification,
        'none': ctx.disable_user_request_all_notification,
        'error': ctx.enable_user_request_error_notification,
        'submission': ctx.enable_user_request_submission_notification,
        'completion': ctx.enable_user_request_completion_notification,
    }
    for n_type in notification_type:
        try:
            f = lookup[n_type]
            f()
            if n_type in ['all', 'none']:
                status = 'enabled' if n_type == 'all' else 'disabled'
                ctx.secho(
                    f'Notifications triggered by requests '
                    f'have been {status}.',
                    fg='green',
                )
            elif n_type in ['error', 'submission', 'completion']:
                ctx.secho(
                    f'Notifications triggered by request {n_type} '
                    f'have been enabled.',
                    fg='green',
                )
        except KeyError:
            pass
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_request_notification_settings()
    columns = ctx.columns or const.COLUMNS_NOT_REQUEST
    ctx.echo(format_output(ctx, [obj], columns=columns, single=True))


@account_notification_set.command('format')
@click.argument('fmt', type=click.Choice(['html', 'text']), required=True)
@pass_context
def account_notification_set_format(ctx: Configuration, fmt):
    """Update notification format where FMT can be html or text."""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.update_user_notification_format(fmt)
    ctx.echo(format_output(ctx, [obj], columns=ctx.columns, single=True))


@account_notification_set.command('method')
@click.argument(
    'method', type=click.Choice(['mail', 'message']), required=True
)
@pass_context
def account_notification_set_method(ctx: Configuration, method):
    """Update notification method where METHOD can be mail or message."""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.update_user_notification_method(method)
    ctx.echo(format_output(ctx, [obj], columns=ctx.columns, single=True))


@account_set.group('mfa', short_help='mfa account settings')
@pass_context
def mfa_set(ctx: Configuration):
    """Set account MFA settings."""
    pass


@mfa_set.command('rm')
@pass_context
def mfa_rm(ctx: Configuration):
    """Disable existing MFA setup."""
    endpoint, username, password = get_endpoint_and_creds(ctx)
    ctx.endpoint = endpoint
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.disable_totp(user=username, password=password)
    ctx.set_defaults()
    columns = ctx.columns or const.COLUMNS_MFA_MIN
    ctx.echo(format_output(ctx, [obj], columns=columns, single=True))
    rm_tk = click.prompt(
        'You should have received an email with a confirmation token. \n'
        'Please, paste the token to continue'
    )
    if rm_tk is not None:
        with ctx.spinner(disable=ctx.debug):
            obj = ctx.disable_totp_confirm(
                token=rm_tk, user=username, password=password
            )
        ctx.echo(format_output(ctx, [obj], columns=columns, single=True))


@mfa_set.command('get-token')
@pass_context
def mfa_request_token(ctx: Configuration):
    """Request TOTP token."""
    endpoint, username, password = get_endpoint_and_creds(ctx)
    try:
        ctx.endpoint = endpoint
        with ctx.spinner(disable=ctx.debug):
            _ = ctx.request_totp(user=username, password=password)
        ctx.secho(
            f'\nVerification code requested {ej_mail}.\n',
            file=sys.stderr,
            fg='green',
        )
    except Exception as ex:
        _LOGGING.error(f'Could not verify TOTP setup: {ex}')


@mfa_set.command('verify')
@click.argument('otp', type=click.STRING, required=False)
@pass_context
def mfa_verify(ctx: Configuration, otp):
    """Verify existing MFA setup."""
    endpoint, username, password = get_endpoint_and_creds(ctx)
    totp = otp or click.prompt(
        'TOTP Code', hide_input=False, type=click.STRING, err=True,
    )

    try:
        ctx.endpoint = endpoint
        with ctx.spinner(disable=ctx.debug):
            obj = ctx.verify_totp(user=username, password=password, otp=totp)
        ctx.secho(
            f'Verification complete: {obj.get("message")}'
            f'{EMOJI_UNICODE.get(":white_heavy_check_mark:")}\n',
            file=sys.stderr,
            fg='green',
        )
    except Exception as ex:
        _LOGGING.error(f'Could not verify TOTP setup: {ex}')


@mfa_set.command('mk')
@click.argument(
    'method',
    type=click.Choice(['EMAIL', 'AUTHENTICATOR', 'SMS']),
    required=True,
)
@click.option(
    '--phone',
    type=click.STRING,
    help='phone number to receive SMS',
    callback=validate_phone_number,
    required=False,
)
@pass_context
def mfa_mk(ctx: Configuration, method: str, phone: str):
    """Enable MFA with Time-based One-Time Password."""
    if method == 'SMS' and phone is None:
        raise click.BadParameter('--phone is required when using SMS')
    endpoint, username, password = get_endpoint_and_creds(ctx)
    ctx.endpoint = endpoint
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.enable_totp(
            user=username, password=password, method=method, phone=phone
        )
    recovery_codes = obj.get('recovery_codes')
    issuer = obj.get('issuer')
    if method == 'AUTHENTICATOR':
        uri = obj.get('uri')
        key = obj.get('key')
        _ = obj.get('image')
        import qrcode

        qr = qrcode.QRCode()
        qr.add_data(uri)
        qr.make(fit=True)
        if click.confirm(
            'Do you have a phone to scan a QR Code to generate TOTP codes?'
        ):
            qr.print_ascii(out=sys.stderr, invert=True)
        if click.confirm('Do you like to display the security key?'):
            ctx.secho(
                'Use the following key if you are unable '
                'to scan the QR Code:\n',
                file=sys.stderr,
                nl=True,
            )
            ctx.secho(
                f'{key}\n', file=sys.stderr, fg='blue', nl=True,
            )
    # print recovery codes.
    if recovery_codes is not None:
        ctx.secho(
            'Recovery codes are used to access your account in \n'
            'the event you cannot get two-factor authentication codes.\n',
            file=sys.stderr,
            nl=True,
        )
        rec_code_txt = '\n'.join(recovery_codes)
        ctx.secho(f'{rec_code_txt}\n', file=sys.stderr, nl=True, fg='blue')
        rec_code_obj = Path(f'{username}_{issuer}_recovery_codes.txt')
        if click.confirm('Would you like to save the codes into a text file?'):
            rec_code_obj.write_text(rec_code_txt)
            click.echo(f'Written {rec_code_obj} with recovery codes.')
    # verify
    if method in ['SMS', 'EMAIL']:
        _ = ctx.request_totp(user=username, password=password)
        ctx.secho(
            f'\nVerification code requested {ej_mail}.\n',
            file=sys.stderr,
            fg='green',
        )
    otp = click.prompt(
        '\nEnter the 6-digit Code to verify enrolment was successful',
        hide_input=False,
        type=click.STRING,
        err=True,
    )
    try:
        obj = ctx.verify_totp(user=username, password=password, otp=otp)
        ctx.secho(
            f'\nVerification complete: {obj.get("message")}'
            f'{EMOJI_UNICODE.get(":white_heavy_check_mark:")}\n',
            file=sys.stderr,
            fg='green',
        )
        success = True
    except Exception as ex:
        success = False
        _LOGGING.error(f'Could not verify TOTP setup: {ex}')
    if success:
        ctx.secho(
            f'You are ready to use the vss-cli '
            f'with MFA via {method} {ej_tada}',
            file=sys.stderr,
            fg='green',
        )



"""Shell plugin for VSS CLI (vss-cli)."""
import logging
import os

import click
from click_repl import repl
from prompt_toolkit.history import FileHistory

from vss_cli import const
from vss_cli.cli import pass_context
from vss_cli.config import Configuration

_LOGGING = logging.getLogger(__name__)


@click.group(
    'shell', short_help='REPL interactive shell.', invoke_without_command=True
)
@click.option(
    '-i',
    '--history',
    type=click.STRING,
    help='File path to save history',
    default=const.DEFAULT_HISTORY,
    envvar='VSS_SHELL_HISTORY',
    required=False,
)
@pass_context
def cli(ctx: Configuration, history):
    """Launch REPL interactive shell."""
    ctx.load_config()
    _message_pfix = 'vss'
    _message_sfix = '> '
    # obtain hostname
    _message = f'{_message_pfix} ({ctx.endpoint_name}) {_message_sfix}'
    welcome = r"""
    __   _____ ___
    \ \ / / __/ __|      API Endpoint: {endpoint}
     \ V /\__ \__ \      Tab-completion & suggestions
      \_/ |___/___/      Prefix external commands with "!"
      CLI {version}       History is saved: {history}

    Exit shell with :exit, :q, :quit, ctrl+d
    """.format(
        version=const.__version__, history=history, endpoint=ctx.endpoint
    )
    ctx.secho(welcome, fg='green')
    # create dir for history
    f_path = os.path.expanduser(history or ctx.history)
    dir_name = os.path.dirname(f_path)

    if not os.path.exists(dir_name):
        _LOGGING.warning(f'Creating {dir_name} for shell history file')
        os.mkdir(dir_name)
        _LOGGING.warning(f'Created {dir_name} for shell history file')
    # run repl
    prompt_kwargs = {'history': FileHistory(f_path), 'message': _message}
    repl(
        click.get_current_context(),
        prompt_kwargs=prompt_kwargs,
        allow_internal_commands=True,
        allow_system_commands=True,
    )



"""Compute Template plugin for VSS CLI (vss-cli)."""
import logging
from typing import List

import click

from vss_cli import autocompletion, const, rel_opts as so
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output
from vss_cli.plugins.compute import cli

_LOGGING = logging.getLogger(__name__)


@cli.group('template', short_help='List virtual machine templates')
@pass_context
def compute_template(ctx):
    """List virtual machine templates."""
    pass


@compute_template.group(
    'rm', help='Delete virtual machine templates', invoke_without_command=True
)
@click.option(
    '-s',
    '--show-info',
    is_flag=True,
    default=False,
    show_default=True,
    help='Show guest info and confirmation.',
)
@click.argument(
    'vm_id',
    type=click.STRING,
    required=True,
    nargs=-1,
    shell_complete=autocompletion.virtual_machines,
)
@so.max_del_opt
@pass_context
def compute_template_rm(
    ctx: Configuration, vm_id: List[str], max_del: int, show_info: bool,
):
    """Delete a list of virtual machine template ids.

    vss-cli compute template rm <name-or-vm-id> <name-or-vm-id> --show-info
    """
    _LOGGING.debug(f'Attempting to remove {vm_id}')
    if len(vm_id) > max_del:
        raise click.BadArgumentUsage(
            'Increase max instance removal with --max-del/-m option'
        )
    # result
    objs = list()
    with ctx.spinner(disable=ctx.debug or show_info):
        for vm in vm_id:
            skip = False
            _vm = ctx.get_vm_by_id_or_name(vm)
            if not _vm:
                _LOGGING.warning(
                    f'Virtual machine {vm} could not be found. Skipping.'
                )
                skip = True
            _LOGGING.debug(f'Found {_vm}')
            moref = _vm[0]['moref']
            if _vm and show_info:
                c_str = const.DEFAULT_VM_DEL_MSG.format(vm=_vm[0])
                confirmation = click.confirm(c_str)
                if not confirmation:
                    _LOGGING.warning(f'Skipping {moref}...')
                    skip = True
            if not skip:
                # request
                objs.append(ctx.delete_template(vm_id=moref))
    # print
    if objs:
        columns = ctx.columns or const.COLUMNS_REQUEST_SUBMITTED
        ctx.echo(
            format_output(ctx, objs, columns=columns, single=len(objs) == 1)
        )
        if ctx.wait_for_requests:
            if len(objs) > 1:
                ctx.wait_for_requests_to(objs, in_multiple=True)
            else:
                ctx.wait_for_request_to(objs[0])
    else:
        _LOGGING.warning('No requests have been submitted.')


@compute_template.command('ls', short_help='List virtual machine templates')
@so.filter_opt
@so.all_opt
@so.page_opt
@so.sort_opt
@so.count_opt
@pass_context
def compute_template_ls(
    ctx: Configuration, filter_by, show_all, sort, page, count
):
    """List virtual machine templates.

    Filter and sort list by any attribute. For example:

    vss-cli compute template ls -f name=like,%vm-name% -f version=like,%13

    Simple name filtering:

    vss-cli compute template ls -f name=%vm-name% -s name=desc

    """
    params = dict(expand=1, sort='name,asc')
    if all(filter_by):
        params['filter'] = ';'.join(filter_by)
    if all(sort):
        params['sort'] = ';'.join(sort)
    # get templates
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_templates(show_all=show_all, per_page=count, **params)
    # including additional attributes?
    columns = ctx.columns or const.COLUMNS_VM_TEMPLATE
    # format output
    output = format_output(ctx, obj, columns=columns)
    # page
    if page:
        click.echo_via_pager(output)
    else:
        ctx.echo(output)



"""Configuration for VSS CLI (vss-cli)."""
from base64 import b64decode, b64encode
import functools
import json
import logging
import os
from pathlib import Path
import platform
import sys
from time import sleep
from typing import (  # noqa: F401
    Any, Callable, Dict, List, Optional, Tuple, Union, cast)

import click
from click_spinner import Spinner, spinner
import jwt
from pick import pick
from pyvss.const import __version__ as pyvss_version
from pyvss.enums import RequestStatus
from pyvss.exceptions import VssError
from pyvss.manager import VssManager
from ruamel.yaml import YAML

import vss_cli.const as const
from vss_cli.data_types import ConfigEndpoint, ConfigFile, ConfigFileGeneral
from vss_cli.exceptions import VssCliError
from vss_cli.helper import (
    bytes_to_str, debug_requests_on, format_output, get_hostname_from_url)
from vss_cli.utils.emoji import EMOJI_UNICODE
from vss_cli.utils.threading import WorkerQueue
from vss_cli.validators import (
    validate_email, validate_phone_number, validate_uuid, validate_vm_moref)
import vss_cli.yaml as yaml

_LOGGING = logging.getLogger(__name__)


class Configuration(VssManager):
    """The configuration context for the VSS CLI."""

    def __init__(self, tk: str = '') -> None:
        """Initialize the configuration."""
        super().__init__(tk)
        self.user_agent = self._default_user_agent(
            extensions=f'pyvss/{pyvss_version}'
        )
        self.verbose = False  # type: bool
        self.default_endpoint_name = None  # type: Optional[str]
        # start of endpoint settings
        self._endpoint = const.DEFAULT_ENDPOINT  # type: str
        self.base_endpoint = self.endpoint  # type: str
        self.endpoint_name = const.DEFAULT_ENDPOINT_NAME
        # end of endpoint settings
        self.history = const.DEFAULT_HISTORY  # type: str
        self.webdav_server = None  # type: Optional[str]
        self._webdav_server = const.DEFAULT_WEBDAV_SERVER  # type: str
        self.username = None  # type: Optional[str]
        self.password = None  # type: Optional[str]
        self.totp = None  # type: Optional[str]
        self.token = None  # type: Optional[str]
        self.timeout = None  # type: Optional[int]
        self._debug = False  # type: Optional[bool]
        self.showexceptions = False  # type: bool
        self.columns = None  # type: Optional[List[Tuple[str, str]]]
        self.columns_width = None  # type: Optional[int]
        self.no_headers = False  # type: Optional[bool]
        self.table_format = None  # type: Optional[str]
        self.sort_by = None  # type: Optional[str]
        self.output = None  # type: Optional[str]
        self.config_path = None  # type: Optional[str]
        self.check_for_updates = None  # type: Optional[bool]
        self.check_for_messages = None  # type: Optional[bool]
        self.wait_for_requests = None  # type: Optional[bool]
        self.config_file = None  # type: Optional[ConfigFile]
        self.spinner = spinner
        self.wait = None  # type: Optional[bool]
        self.moref = None  # type: Optional[str]
        self.unit = None  # type: Optional[str, int]
        self.payload_options = {}  # type: Optional[Dict]
        self.tmp = None  # type: Optional[Any]

    def set_dry_run(self, val: bool) -> None:
        """Set dry_run value."""
        if val is True:
            if self.output not in ['json', 'yaml']:
                self.output = 'json'
            self.wait = not bool(val)
            self.dry_run = bool(val)

    @property
    def debug(self) -> bool:
        """Return debug status."""
        return self._debug

    @debug.setter
    def debug(self, value: bool) -> None:
        """Set on debug_requests if True."""
        if value:
            debug_requests_on()
        self._debug = value

    @property
    def endpoint(self) -> str:
        """Return endpoint value."""
        return self._endpoint

    @endpoint.setter
    def endpoint(self, value: str) -> None:
        """Rebuild API endpoints."""
        self._endpoint = value
        self.base_endpoint = value
        self.api_endpoint = f'{value}/v2'
        self.token_endpoint = f'{value}/auth/request-token'
        self.tf_endpoint = f'{value}/tf'
        if value:
            self.endpoint_name = get_hostname_from_url(
                value, const.DEFAULT_HOST_REGEX
            )

    def set_defaults(self) -> None:
        """Set default configuration settings."""
        _LOGGING.debug('Setting default configuration.')
        for setting, default in const.DEFAULT_SETTINGS.items():
            if getattr(self, setting) is None:
                setattr(self, setting, default)
        _LOGGING.debug(self)

    def get_token(
        self,
        user: Optional[str] = '',
        password: Optional[str] = '',
        otp: Optional[str] = None,
    ) -> str:
        """Generate token and returns value."""
        self.api_token = super().get_token(user, password, otp)
        _LOGGING.debug(f'Token generated successfully: {self.api_token}')
        return self.api_token

    def update_endpoints(self, endpoint: str = '') -> None:
        """Rebuild API endpoints."""
        self.base_endpoint = endpoint
        self.api_endpoint = f'{endpoint}/v2'
        self.token_endpoint = f'{endpoint}/auth/request-token'
        self.tf_endpoint = f'{endpoint}/tf'

    def echo(self, msg: str, *args: Optional[Any]) -> None:
        """Put content message to stdout."""
        self.log(msg, *args)

    def log(  # pylint: disable=no-self-use
        self, msg: str, *args: Optional[str]
    ) -> None:  # pylint: disable=no-self-use
        """Log a message to stdout."""
        if args:
            msg %= args
        click.echo(msg, file=sys.stdout)

    def secho(self, msg: str, *args: Optional[Any], **kwargs) -> None:
        """Put content message to stdout with style."""
        self.slog(msg, *args, **kwargs)

    def slog(  # pylint: disable=no-self-use
        self, msg: str, *args: Optional[str], **kwargs
    ) -> None:  # pylint: disable=no-self-use
        """Log a message to stdout with style."""
        file = sys.stdout
        if args:
            msg %= args
        if 'file' in kwargs:
            file = kwargs['file']
            del kwargs['file']
        click.secho(msg, file=file, **kwargs)

    def vlog(self, msg: str, *args: Optional[str]) -> None:
        """Log a message only if verbose is enabled."""
        if self.verbose:
            self.log(msg, *args)

    def __repr__(self) -> str:
        """Return the representation of the Configuration."""
        view = {
            "endpoint": self.endpoint,
            "default_endpoint_name": self.default_endpoint_name,
            "endpoint_name": self.endpoint_name,
            "access_token": 'yes' if self.token is not None else 'no',
            "user": 'yes' if self.username is not None else 'no',
            "user_password": 'yes' if self.password is not None else 'no',
            "tf_enabled": 'yes' if self.tf_enabled is not None else 'no',
            "output": self.output,
            "timeout": self.timeout,
            "debug": self.debug,
            "verbose": self.verbose,
            "wait": self.wait,
            "dry_run": self.dry_run,
            "webdav_server": self.webdav_server,
        }

        return f"<Configuration({view})"

    def auto_output(self, auto_output: str) -> str:
        """Configure output format."""
        if self.output == "auto":
            if auto_output == 'data':
                auto_output = const.DEFAULT_RAW_OUTPUT
            _LOGGING.debug(f"Setting auto-output to: {auto_output}")
            self.output = auto_output
        return self.output

    @staticmethod
    def _default_user_agent(
        name: str = const.PACKAGE_NAME,
        version: str = const.__version__,
        extensions: str = '',
    ) -> str:
        """Set default user agent."""
        # User-Agent:
        # <product>/<version> (<system-information>)
        # <platform> (<platform-details>) <extensions>
        user_agent = (
            f'{name}/{version} '
            f'({platform.system()}/{platform.release()}) '
            f'Python/{platform.python_version()} '
            f'({platform.platform()}) {extensions}'
        )
        return user_agent

    def set_credentials(
        self,
        username: str,
        password: str,
        token: str,
        endpoint: str,
        name: str,
    ) -> None:
        """Set credentials.

        Username, password, Token, endpoint name.
        Useful for configuration purposes.
        """
        self.username = username
        self.password = password
        self.api_token = token
        self.token = token
        self.endpoint = endpoint
        self.endpoint_name = name
        return

    def load_profile(
        self, endpoint: str = None
    ) -> Tuple[Optional[ConfigEndpoint], Optional[str], Optional[str]]:
        """Load profile from configuration file."""
        username, password = None, None
        # load from
        config_endpoint = self.config_file.get_endpoint(endpoint)
        if config_endpoint:
            # get auth attr
            auth = config_endpoint[0].auth
            # get token attr
            token = config_endpoint[0].token
            if not auth or not token:
                _LOGGING.warning('Invalid configuration endpoint found.')
            else:
                auth_enc = auth.encode()
                credentials_decoded = b64decode(auth_enc)
                # get user/pass
                username, password = credentials_decoded.split(b':')
            return (
                config_endpoint[0],
                bytes_to_str(username),
                bytes_to_str(password),
            )
        else:
            return None, bytes_to_str(username), bytes_to_str(password)

    def load_config_file(
        self, config: Union[Path, str] = None
    ) -> Optional[ConfigFile]:
        """Load raw configuration file and return ConfigFile object."""
        raw_config = self.load_raw_config_file(config=config)
        self.config_file = ConfigFile.from_json(raw_config)
        return self.config_file

    def load_raw_config_file(
        self, config: Optional[Union[Path, str]] = None
    ) -> Optional[str]:
        """Load raw configuration file from path."""
        config_file = config or self.config_path
        try:
            if isinstance(config_file, str):
                config_file = Path(config_file)
            cfg_dict = self.yaml_load(config_file)
            return json.dumps(cfg_dict)
        except ValueError as ex:
            _LOGGING.error(f'Error loading configuration file: {ex}')
            raise VssCliError(
                'Invalid configuration file.'
                'Run "vss-cli configure mk" or '
                '"vss-cli configure upgrade" to upgrade '
                'legacy configuration.'
            )

    def load_config(
        self, validate: bool = True, spinner_cls: Optional[Spinner] = None
    ) -> Optional[Tuple[str, str, str]]:
        """Load configuration and validate.

        Load configuration either from previously set
        ``username`` and ``password`` or ``token``.
        """
        try:
            # input configuration check
            if self.token or (self.username and self.password):
                # setting defaults if required
                self.set_defaults()
                _LOGGING.debug('Loading from input')
                # don't load config_path file
                if self.token:
                    _LOGGING.debug('Checking token')
                    # set api token
                    self.api_token = self.token
                    return self.username, self.password, self.api_token
                elif self.username and self.password:
                    _LOGGING.debug('Checking user/pass to generate token')
                    # generate a new token - won't save
                    _LOGGING.warning(
                        'A new token will be generated but not persisted. '
                        'Consider running command "configure mk" to save your '
                        'credentials.'
                    )
                    self.get_token(self.username, self.password, self.totp)
                    return self.username, self.password, self.api_token
                else:
                    raise VssCliError(
                        'Environment variables error. Please, verify '
                        'VSS_TOKEN or VSS_USER and VSS_USER_PASS'
                    )
            else:
                cfg_path = Path(self.config_path)
                _LOGGING.debug(
                    f'Loading configuration file: {self.config_path}'
                )
                if cfg_path.is_file():
                    # load configuration file from json string into class
                    self.config_file = self.load_config_file(config=cfg_path)
                    # general area
                    if self.config_file.general:
                        _LOGGING.debug(
                            f'Loading general settings from {self.config_path}'
                        )
                        # set config_path defaults
                        for setting in const.GENERAL_SETTINGS:
                            try:
                                # check if setting hasn't been set
                                # by input or env
                                # which overrides configuration file
                                if getattr(self, setting) is None:
                                    setattr(
                                        self,
                                        setting,
                                        getattr(
                                            self.config_file.general, setting
                                        ),
                                    )
                                else:
                                    _LOGGING.debug(
                                        f'Prioritizing {setting} from '
                                        f'command line input.'
                                    )
                            except KeyError as ex:
                                _LOGGING.warning(
                                    f'Could not load general setting'
                                    f' {setting}: {ex}'
                                )
                        # printing out
                        _LOGGING.debug(f"General settings loaded: {self}")
                    # load preferred endpoint from file if any
                    if self.config_file.endpoints:
                        _LOGGING.debug(
                            f'Loading endpoint settings from '
                            f'{self.config_path}'
                        )
                        _LOGGING.debug(
                            f'Looking for endpoint={self.endpoint},'
                            f' default_endpoint_name='
                            f'{self.default_endpoint_name}'
                        )
                        # 1. provided by input
                        if self.endpoint:
                            msg = (
                                f'Cloud not find endpoint provided by '
                                f'input {self.endpoint}. \n'
                            )
                            # load endpoint from endpoints
                            ep, usr, pwd = self.load_profile(self.endpoint)
                        # 2. provided by configuration file
                        #    (default_endpoint_name)
                        elif self.default_endpoint_name:
                            msg = (
                                f'Could not find default endpoint '
                                f'{self.default_endpoint_name}. \n'
                            )
                            # load endpoint from endpoints
                            ep, usr, pwd = self.load_profile(
                                self.default_endpoint_name
                            )
                        # 3. fallback to default settings
                        else:
                            msg = (
                                f"Invalid endpoint {self.endpoint_name} "
                                f"configuration. \n"
                            )
                            ep, usr, pwd = self.load_profile(
                                self.endpoint_name
                            )
                        # check valid creds
                        if not (usr and pwd or getattr(ep, 'token', None)):
                            _LOGGING.warning(msg)
                            default_endpoint = const.DEFAULT_ENDPOINT_NAME
                            _LOGGING.warning(
                                f'Falling back to {default_endpoint}'
                            )
                            (ep, usr, pwd,) = self.load_profile(  # NOQA: E501
                                endpoint=const.DEFAULT_ENDPOINT_NAME
                            )
                        # set config_path data
                        self.set_credentials(
                            usr, pwd, ep.token, ep.url, ep.name,
                        )
                        if validate:
                            # last check cred
                            creds = self.username and self.password
                            if not (creds or self.api_token):
                                raise VssCliError(
                                    'Run "vss-cli configure mk" to add '
                                    'endpoint to configuration file or '
                                    '"vss-cli configure upgrade" to upgrade '
                                    'legacy configuration.'
                                )
                            _LOGGING.debug(
                                f'Loaded from file: {self.endpoint_name}: '
                                f'{self.endpoint}:'
                                f': {self.username}'
                            )
                            try:
                                self.whoami()
                                _LOGGING.debug('Token validated successfully.')
                            except Exception as ex:
                                self.vlog(str(ex))
                                _LOGGING.debug('Generating a new token')
                                self.api_token = self._get_token_with_mfa(
                                    spinner_cls=spinner_cls
                                )
                                endpoint = self._create_endpoint_config(
                                    token=self.api_token
                                )
                                self.write_config_file(new_endpoint=endpoint)
                                # check for updates
                                if self.check_for_updates:
                                    self.check_available_updates()
                                # check for unread messages
                                if self.check_for_messages:
                                    self.check_unread_messages()
                        return self.username, self.password, self.api_token
                else:
                    self.set_defaults()
            raise VssCliError(
                'Invalid configuration. Please, run '
                '"vss-cli configure mk" to initialize configuration, or '
                '"vss-cli configure upgrade" to upgrade legacy '
                'configuration.'
            )
        except Exception as ex:
            raise VssCliError(str(ex))

    def check_available_updates(self) -> None:
        """Check available update from PyPI."""
        try:
            _LOGGING.debug('Checking for available updates.')
            cmd_bin = sys.executable
            # create command with the right exec
            pip_cmd = f'{cmd_bin} -m pip list --outdated'.split(None)
            from subprocess import Popen, PIPE

            p = Popen(pip_cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
            out, err = p.communicate()
            out_decoded = out.decode('utf-8')
            # verify if package name is in outdated string
            pkg_name = const.PACKAGE_NAME.replace('_', '-')
            if pkg_name in out_decoded:
                lines = out_decoded.split('\n')
                pkg_line = [line for line in lines if pkg_name in line]
                if pkg_line:
                    pkg_line = pkg_line.pop()
                    pkg, current, latest, pkgn = pkg_line.split(None)
                    self.secho(
                        f'Update available {current} -> {latest} '
                        f'{EMOJI_UNICODE.get(":upwards_button:")}.',
                        file=sys.stderr,
                        fg='green',
                        nl=False,
                    )
                    self.secho(' Run ', file=sys.stderr, fg='green', nl=False)
                    self.secho(
                        'vss-cli upgrade', file=sys.stderr, fg='red', nl=False
                    )
                    self.secho(
                        ' to install latest. \n', file=sys.stderr, fg='green'
                    )
            else:
                self.secho(
                    f'Running latest version {const.__version__} '
                    f'{EMOJI_UNICODE.get(":white_heavy_check_mark:")}\n',
                    file=sys.stderr,
                    fg='green',
                )
        except Exception as ex:
            _LOGGING.error(f'Could not check for updates: {ex}')

    def check_unread_messages(self) -> None:
        """Check unread API messages."""
        try:
            _LOGGING.debug('Checking for unread messages')
            messages = self.get_user_messages(
                filter='status,eq,Created', per_page=100
            )
            n_messages = len(messages)
            if messages:
                self.secho(
                    f'You have {n_messages} unread messages '
                    f'{EMOJI_UNICODE.get(":envelope_with_arrow:")} ',
                    file=sys.stderr,
                    fg='green',
                    nl=False,
                )
                self.secho('Run ', file=sys.stderr, fg='green', nl=False)
                self.secho(
                    'vss-cli message ls -f status=Created',
                    file=sys.stderr,
                    fg='red',
                    nl=False,
                )
                self.secho(
                    ' to list unread messages.', file=sys.stderr, fg='green'
                )
            else:
                _LOGGING.debug('No messages with Created status')
        except Exception as ex:
            _LOGGING.error(f'Could not check for messages: {ex}')

    def _get_token_with_mfa(
        self,
        token: Optional[str] = None,
        spinner_cls: Optional[Spinner] = None,
    ):
        """Get token with MFA."""
        try:
            token = token or self.get_token(
                self.username, self.password, self.totp
            )
        except VssError as ex:
            if (
                500 > ex.http_code > 399
                and 'InvalidParameterValue: otp' in ex.message
            ):
                # try MFA Auth
                try:
                    _LOGGING.debug('Requesting a new timed one-time password')
                    _ = self.request_totp(self.username, self.password)
                except Exception as exc:
                    _LOGGING.warning(f'Requesting totp: {exc}')
                    pass
                if spinner_cls is not None:
                    spinner_cls.stop()
                self.totp = click.prompt(
                    'MFA enabled. Provide Timed One-Time Password'
                )
                if spinner_cls is not None:
                    spinner_cls.start()
                try:
                    token = token or self.get_token(
                        self.username, self.password, self.totp
                    )
                except Exception as exc:
                    _LOGGING.warning(f'Could not generate new token: {exc}')
                    raise exc
            else:
                raise ex
        return token

    def _create_endpoint_config(self, token: str = None) -> ConfigEndpoint:
        """Create endpoint configuration for a given token.

        Token might be ``None`` and will generate a new one
        using ``username`` and ``password``.
        """
        token = self._get_token_with_mfa(token=token)
        # encode or save
        username = (
            self.username
            if isinstance(self.username, bytes)
            else self.username.encode()
        )
        password = (
            self.password
            if isinstance(self.password, bytes)
            else self.password.encode()
        )
        credentials = b':'.join([username, password])
        auth = b64encode(credentials).strip().decode('utf-8')
        # decode jwt to verify if otp is enabled.
        payload = jwt.decode(
            self.api_token, options=dict(verify_signature=False)
        )
        endpoint_cfg = {
            'url': self.base_endpoint,
            'name': self.endpoint_name,
            'auth': auth,
            'token': self.api_token,
            'tf_enabled': payload.get('otp', False),
        }
        ep_cfg = ConfigEndpoint.from_json(json.dumps(endpoint_cfg))
        _LOGGING.debug(f'Configuration endpoint created: {ep_cfg}')
        return ep_cfg

    def load_config_template(self) -> ConfigFile:
        """Load configuration from template."""
        # load template in case it fails
        cfg_default_path = Path(const.DEFAULT_CONFIG_TMPL)
        with cfg_default_path.open() as f:
            config_tmpl = yaml.load_yaml(self.yaml(), f)
            raw_config_tmpl = json.dumps(config_tmpl)
            config_file_tmpl = ConfigFile.from_json(raw_config_tmpl)
        return config_file_tmpl

    def write_config_file(
        self,
        new_config_file: Optional[ConfigFile] = None,
        new_endpoint: Optional[ConfigEndpoint] = None,
        config_general: Optional[ConfigFileGeneral] = None,
    ) -> bool:
        """Create or update configuration endpoint section."""
        # load template in case it fails
        config_file_tmpl = self.load_config_template()
        try:
            _LOGGING.debug(f'Writing configuration file: {self.config_path}')
            cfg_path = Path(self.config_path)
            # validate if file exists
            if cfg_path.is_file():
                with cfg_path.open(mode='r+') as fp:
                    try:
                        _conf_dict = yaml.load_yaml(self.yaml(), fp)
                        raw_config = json.dumps(_conf_dict)
                        config_file = ConfigFile.from_json(raw_config)
                    except (ValueError, TypeError) as ex:
                        _LOGGING.warning(f'Invalid config_path file: {ex}')
                        if click.confirm(
                            'An error occurred loading the '
                            'configuration file. '
                            'Would you like to recreate it?',
                            err=True,
                        ):
                            config_file = config_file_tmpl
                        else:
                            return False
                    if new_config_file:
                        config_file.general = new_config_file.general
                        config_file.update_endpoints(
                            *new_config_file.endpoints
                        )
                    # update general config_path if required
                    if config_general:
                        config_file.general = config_general
                    # update endpoint if required
                    if new_endpoint:
                        # update endpoint
                        config_file.update_endpoint(new_endpoint)
                    # dumping and loading
                    _conf_dict = json.loads(config_file.to_json())
                    fp.seek(0)
                    yaml.dump_yaml(self.yaml(), _conf_dict, stream=fp)
                    fp.truncate()
                _LOGGING.debug(
                    f'Configuration file {self.config_path} has been updated'
                )
            else:
                if new_config_file:
                    f_type = 'Config file'
                    config_file_dict = json.loads(new_config_file.to_json())
                else:
                    # New configuration file. A new endpoint must be configured
                    f_type = 'Default template'
                    config_endpoint = (
                        new_endpoint or self._create_endpoint_config()
                    )
                    config_file_tmpl.update_endpoint(config_endpoint)
                    # load and dump
                    config_file_dict = json.loads(config_file_tmpl.to_json())
                # write file
                with cfg_path.open(mode='w') as fp:
                    yaml.dump_yaml(self.yaml(), config_file_dict, stream=fp)
                _LOGGING.debug(
                    f'New {f_type} has been written to {self.config_path}.'
                )
        except OSError as e:
            raise Exception(
                f'An error occurred writing ' f'configuration file: {e}'
            )
        return True

    def configure(
        self,
        username: str,
        password: str,
        endpoint: str,
        replace: Optional[bool] = False,
        endpoint_name: Optional[str] = None,
    ) -> bool:
        """Configure endpoint with provided settings."""
        self.username = username
        self.password = password
        # update instance endpoints if provided
        self.endpoint = endpoint
        if endpoint_name:
            self.endpoint_name = endpoint_name
        # directory available
        cfg_path = Path(self.config_path)
        if not cfg_path.parent.is_dir():
            cfg_path.parent.mkdir()
        # config_path file
        if cfg_path.is_file():
            try:
                self.config_file = self.load_config_file()
                # load credentials by endpoint_name
                (
                    config_endpoint,
                    e_username,
                    e_password,
                ) = self.load_profile(  # NOQA: E501c
                    endpoint=self.endpoint_name
                )
                # profile does not exist
                if not (e_username and e_password and config_endpoint.token):
                    self.echo(
                        f'Endpoint {self.endpoint_name} not found. '
                        f'Creating...'
                    )
                    endpoint_cfg = self._create_endpoint_config()
                    self.write_config_file(new_endpoint=endpoint_cfg)
                # profile exists
                elif e_username and e_password and config_endpoint.token:
                    username = bytes_to_str(e_username) if e_username else ''
                    confirm = replace or click.confirm(
                        f"Would you like to replace existing configuration?\n "
                        f"{self.endpoint_name}:"
                        f"{username}: {config_endpoint.url}",
                        err=True,
                    )
                    if confirm:
                        endpoint_cfg = self._create_endpoint_config()
                        self.write_config_file(new_endpoint=endpoint_cfg)
                    else:
                        return False
            except Exception as ex:
                _LOGGING.warning(f'Invalid config_path file: {ex}')
                confirm = click.confirm(
                    'An error occurred loading the '
                    'configuration file. '
                    'Would you like to recreate it?',
                    err=True,
                )
                if confirm:
                    endpoint_cfg = self._create_endpoint_config()
                    return self.write_config_file(new_endpoint=endpoint_cfg)
                else:
                    return False
            # feedback
            self.echo(
                f'Successfully configured credentials for ' f'{self.endpoint}.'
            )
        else:
            endpoint_cfg = self._create_endpoint_config()
            self.write_config_file(new_endpoint=endpoint_cfg)
        return True

    @staticmethod
    def _filter_objects_by_attrs(
        value: str, objects: List[dict], attrs: List[Tuple[Any, Any]]
    ) -> List[Any]:
        """Filter objects by a given `value` based on attributes.

        Attributes may be a list of tuples with attribute name, type.

        :param value: value to filter
        :param objects: list of dictionaries
        :param attrs: list of tuple of attribute name, type
        :return:
        """
        _objs = []
        for attr in attrs:
            attr_name = attr[0]
            attr_type = attr[1]
            try:
                if attr_type in [str]:
                    f = filter(
                        lambda i: attr_type(value).lower()
                        in i[attr_name].lower(),
                        objects,
                    )
                elif attr_type in [int]:
                    f = filter(
                        lambda i: attr_type(value) == i[attr_name], objects
                    )
                else:
                    f = filter(
                        lambda i: attr_type(value) in i[attr_name], objects
                    )
                # cast list
                _objs = list(f)
            except ValueError as ex:
                _LOGGING.debug(f'{value} ({type(value)}) error: {ex}')

            if _objs:
                break
        return _objs

    @staticmethod
    def pick(objects: List[Dict], options=None, indicator='=>'):
        """Show a ``picker`` for a list of dicts."""
        count = len(objects)
        msg = f"Found {count} matches. Please select one:"
        sel, index = pick(
            title=msg, indicator=indicator, options=options or objects
        )
        return [objects[index]]

    def get_vskey_stor(self, **kwargs) -> bool:
        """Create WebDav client to interact with remote CrushFTP."""
        try:
            from webdav3 import client as wc
        except ImportError:
            raise VssCliError(
                'webdavclient3 dependency not found. '
                'try running "pip install vss-cli[stor]"'
            )

        options = dict(
            webdav_login=self.username,
            webdav_password=self.password,
            webdav_hostname=self.webdav_server,
            verbose=self.verbose,
        )
        self.vskey_stor = wc.Client(options=options)
        return self.vskey_stor.valid()

    def get_vm_by_id_or_name(self, vm_id: str, silent=False) -> Optional[List]:
        """Get virtual machine by identifier or name."""
        is_moref = validate_vm_moref('', '', vm_id)
        is_uuid = validate_uuid('', '', vm_id)
        _LOGGING.debug(f'is_moref={is_moref}, is_uuid={is_uuid}')
        if is_moref or is_uuid:
            if is_moref:
                attr = 'moref'
            else:
                attr = 'uuid'
            filters = f'{attr},eq,{vm_id}'
            v = self.get_vms(filter=filters)
            if not v:
                # try template
                v = self.get_templates(filter=filters)
                if not v:
                    if silent:
                        return None
                    else:
                        raise click.BadArgumentUsage(
                            f'vm id {vm_id} could not be found'
                        )
            return v
        else:
            _LOGGING.debug(f'not a moref or uuid {vm_id}')
            # If it's a value error, then the string
            # is not a valid hex code for a UUID.
            # get vm by name
            g_vms = self.get_vms(per_page=3000)
            vm_id = vm_id.lower()
            v = list(filter(lambda i: vm_id in i['name'].lower(), g_vms))
            if not v:
                # try templates:
                g_tmpls = self.get_templates(per_page=2500)
                v = list(filter(lambda i: vm_id in i['name'].lower(), g_tmpls))
                if not v:
                    raise click.BadParameter(f'{vm_id} could not be found')
            v_count = len(v)
            if v_count > 1:
                msg = f"Found {v_count} matches. Please select one:"
                sel, index = pick(
                    title=msg,
                    indicator='=>',
                    options=[
                        f"({i['moref']}) {i['folder']['path']} > {i['name']}"
                        for i in v
                    ],
                )
                return [v[index]]
            return v

    def get_vm_snapshot_by_id_name_or_desc(
        self, vm_id: str, id_name_or_desc: str
    ) -> List[Dict]:
        """Get vm snapshot by id name or description."""
        snapshots = self.get_vm_snapshots(vm_id)
        attributes = [('id', int), ('name', str), ('description', str)]
        objs = self._filter_objects_by_attrs(
            id_name_or_desc, snapshots, attributes
        )
        if not objs:
            raise click.BadParameter(f'{id_name_or_desc} could not be found')
        d_count = len(objs)
        if d_count > 1:
            return self.pick(
                objs,
                options=[f"{i['name']} ({i['description']})" for i in objs],
            )
        return objs

    def get_domain_by_name_or_moref(self, name_or_moref: str) -> List[Dict]:
        """Get domain by name or mo reference."""
        g_domains = self.get_domains()
        attributes = [('name', str), ('moref', str)]
        objs = self._filter_objects_by_attrs(
            name_or_moref, g_domains, attributes
        )
        if not objs:
            raise click.BadParameter(f'{name_or_moref} could not be found')
        d_count = len(objs)
        if d_count > 1:
            return self.pick(
                objs, options=[f"{i['name']} ({i['moref']})" for i in objs]
            )
        return objs

    def get_network_by_name_or_moref(self, name_or_moref: str) -> List[Dict]:
        """Get network by name or mo reference."""
        g_networks = self.get_networks(
            sort='name,desc', show_all=True, per_page=500
        )
        attributes = [('name', str), ('moref', str)]
        objs = self._filter_objects_by_attrs(
            name_or_moref, g_networks, attributes
        )
        if not objs:
            raise click.BadParameter(f'{name_or_moref} could not be found')
        net_count = len(objs)
        if net_count > 1:
            return self.pick(
                objs, options=[f"{i['name']} ({i['moref']})" for i in objs]
            )
        return objs

    def get_folder_by_name_or_moref_path(
        self, name_moref_path: str, silent: bool = False
    ) -> List[Dict]:
        """Get domain by name or mo reference."""
        g_folders = self.get_folders(
            sort='path,desc', show_all=True, per_page=500
        )
        # search by name or moref
        attributes = [('name', str), ('path', str), ('moref', str)]
        objs = self._filter_objects_by_attrs(
            name_moref_path, g_folders, attributes
        )
        if not objs and not silent:
            raise click.BadParameter(f'{name_moref_path} could not be found')
        obj_count = len(objs)
        if obj_count > 1:
            return self.pick(
                objs, options=[f"{i['path']} ({i['moref']})" for i in objs]
            )
        return objs

    def get_os_by_name_or_guest(self, name_or_guest: str) -> List[Dict]:
        """Get operating system by name, ``guest_id`` or ``full_name``."""
        g_os = self.get_os(
            sort='guestFullName,desc', show_all=True, per_page=500
        )
        attributes = [('id', int), ('guest_id', str), ('full_name', str)]
        objs = self._filter_objects_by_attrs(name_or_guest, g_os, attributes)
        if not objs:
            raise click.BadParameter(f'{name_or_guest} could not be found')
        o_count = len(objs)
        if o_count > 1:
            return self.pick(
                objs,
                options=[f"{i['full_name']} ({i['guest_id']})" for i in objs],
            )
        return objs

    def get_vss_service_by_name_label_or_id(
        self, name_label_or_id: Union[str, int]
    ) -> List[Dict]:
        """Get service by name label or identifier."""
        vss_services = self.get_vss_services(
            sort='label,desc', show_all=True, per_page=200
        )
        attributes = [('id', int), ('label', str), ('name', str)]
        objs = self._filter_objects_by_attrs(
            name_label_or_id, vss_services, attributes
        )
        # check if there's no ref
        if not objs:
            raise click.BadParameter(f'{name_label_or_id} could not be found')
        # count for dup results
        o_count = len(objs)
        if o_count > 1:
            return self.pick(objs, options=[f"{i['label']}" for i in objs])
        return objs

    def get_vss_groups_by_name_desc_or_id(
        self, name_desc_or_id: Union[str, int]
    ) -> List[Dict]:
        """Get groups by name, description or identifier."""
        vss_groups = self.get_user_groups(
            sort='name,desc', show_all=True, per_page=100
        )
        attributes = [('id', int), ('name', str), ('description', str)]
        objs = self._filter_objects_by_attrs(
            name_desc_or_id, vss_groups, attributes
        )
        # check if there's no ref
        if not objs:
            raise click.BadParameter(f'{name_desc_or_id} could not be found')
        # count for dup results
        o_count = len(objs)
        if o_count > 1:
            return self.pick(
                objs,
                options=[
                    f"{i['name']} ({i['id']}): {i['description'][:40]}... "
                    for i in objs
                ],
            )
        return objs

    def _get_images_by_name_path_or_id(
        self, f: Callable, name_or_path_or_id: Union[int, str]
    ) -> List[Dict]:
        """Get images by name path or identifier."""
        g_img = f(show_all=True, per_page=500)
        attributes = [('id', int), ('name', str), ('path', str)]
        objs = self._filter_objects_by_attrs(
            name_or_path_or_id, g_img, attributes
        )
        # check if there's no ref
        if not objs:
            raise click.BadParameter(
                f'{name_or_path_or_id} could not be found'
            )
        # count for dup results
        o_count = len(objs)
        if o_count > 1:
            return self.pick(objs, options=[f"{i['name']}" for i in objs])
        return objs

    def get_vmdk_by_name_path_or_id(
        self, name_or_path_or_id: Union[str, int]
    ) -> List[Any]:
        """Get vmdk by name, path or id."""
        return self._get_images_by_name_path_or_id(
            self.get_user_vmdks, name_or_path_or_id
        )

    def get_floppy_by_name_or_path(
        self, name_or_path_or_id: Union[str, int]
    ) -> List[Dict]:
        """Get Floppy image by name, path or identifier."""
        return self._get_images_by_name_path_or_id(
            self.get_floppies, name_or_path_or_id
        )

    def get_iso_by_name_or_path(
        self, name_or_path_or_id: Union[str, int]
    ) -> List[Any]:
        """Get ISO image by name, path or identifier."""
        return self._get_images_by_name_path_or_id(
            self.get_isos, name_or_path_or_id
        )

    def get_clib_deployable_item_by_name_or_id_path(
        self, name_or_id_or_path: Union[str, int]
    ):
        """Get content library deployable items."""
        items = self.get_content_library_items(
            show_all=True,
            sort='path,desc',
            per_page=500,
            filter="type,in,OVF,VM_TEMPLATE",
        )
        attrs = [('id', int), ('name', str)]
        objs = self._filter_objects_by_attrs(name_or_id_or_path, items, attrs)
        # check if there's no ref
        if not objs:
            raise click.BadParameter(
                f'{name_or_id_or_path} could not be found'
            )
        # count for dup results
        o_count = len(objs)
        if o_count > 1:
            return self.pick(objs, options=[f"{i['name']}" for i in objs])
        return objs

    def get_vm_image_by_name_or_id_path(
        self, name_or_path_or_id: Union[str, int]
    ) -> List[Any]:
        """Get VM image by name, path or identifier."""
        return self._get_images_by_name_path_or_id(
            self.get_images, name_or_path_or_id
        )

    def _get_types_by_name(self, name: Union[str, int], types_f, attrs=None):
        g_types = types_f(only_type=False)
        attributes = attrs or [('type', str)]
        objs = self._filter_objects_by_attrs(str(name), g_types, attributes)
        # check if there's no ref
        if not objs:
            raise click.BadParameter(f'{name} could not be found')
        # count for dup results
        o_count = len(objs)
        if o_count > 1:
            return self.pick(
                objs,
                options=[
                    f"{i['type']} - {i['description'][:100]}..." for i in objs
                ],
            )
        return objs

    def get_vm_scsi_type_by_name(self, name: Union[str, int]):
        """Get SCSI type by name."""
        return self._get_types_by_name(
            name, self.get_supported_scsi_controllers
        )

    def get_vm_scsi_sharing_by_name(self, name: Union[str, int]):
        """Get SCSI sharing by name."""
        return self._get_types_by_name(name, self.get_supported_scsi_sharing)

    def get_vm_disk_backing_mode_by_name(self, name: Union[str, int]):
        """Get Disk Backing Mode by name."""
        return self._get_types_by_name(
            name, self.get_supported_disk_backing_modes
        )

    def get_vm_disk_backing_sharing_by_name(self, name: Union[str, int]):
        """Get Disk Sharing Mode by name."""
        return self._get_types_by_name(name, self.get_supported_disk_sharing)

    def get_vm_firmware_by_type_or_desc(self, name: Union[str, int]):
        """Get VM firmware by name."""
        return self._get_types_by_name(
            name,
            self.get_supported_firmware_types,
            attrs=[('type', str), ('description', str)],
        )

    def get_vm_nic_type_by_name(self, name: Union[str, int]):
        """Get VM NIC type by name."""
        g_types = self.get_supported_nic_types(only_type=False)
        attributes = [('type', str)]
        objs = self._filter_objects_by_attrs(name, g_types, attributes)
        # check if there's no ref
        if not objs:
            raise click.BadParameter(f'{name} could not be found')
        # count for dup results
        o_count = len(objs)
        if o_count > 1:
            return self.pick(
                objs,
                options=[
                    f"{i['type']} - {i['description'][:100]}..." for i in objs
                ],
            )
        return objs

    def get_cli_spec_from_api_spec(
        self, payload: dict, template: dict
    ) -> Dict:
        """Get CLI specification from API specification."""
        os_q = self.get_os(filter=f"guest_id,eq,{payload.get('os')}")
        machine_os = os_q[0]['full_name'] if os_q else payload.get('os')
        fo_q = self.get_folder(payload.get('folder'))
        machine_folder = fo_q['path'] if fo_q else payload.get('folder')
        template['built'] = payload.get('built_from')
        template['machine']['name'] = payload.get('name')
        template['machine']['os'] = machine_os
        template['machine']['cpu'] = payload.get('cpu')
        template['machine']['memory'] = payload.get('memory')
        template['machine']['folder'] = machine_folder
        template['machine']['disks'] = payload.get('disks')
        template['networking']['interfaces'] = [
            {
                'network': self.get_network(v['network'])['name'],
                'type': v['type'],
            }
            for v in payload.get('networks')
        ]
        template['metadata']['client'] = payload.get('client')
        template['metadata']['description'] = payload.get('description')
        template['metadata']['usage'] = payload.get('usage')
        template['metadata']['inform'] = payload.get('inform')
        template['metadata']['admin'] = {
            'name': payload.get('admin_name'),
            'email': payload.get('admin_email'),
            'phone': payload.get('admin_phone'),
        }
        template['metadata']['vss_service'] = payload.get('vss_service')
        template['metadata']['vss_options'] = payload.get('vss_options')
        return template

    def get_api_spec_from_cli_spec(self, payload: dict, built: str) -> Dict:
        """Get API specification from CLI specification."""
        try:
            spec_payload = dict()
            # sections
            machine_section = payload['machine']
            networking_section = payload['networking']
            metadata_section = payload['metadata']
            if built == 'os_install':
                # machine section parse and update
                spec_payload.update(machine_section)
                # replace with valid values
                spec_payload['os'] = self.get_os_by_name_or_guest(
                    machine_section['os']
                )[0]['guest_id']
                spec_payload['iso'] = self.get_iso_by_name_or_path(
                    machine_section['iso']
                )[0]['path']
                # folder
                spec_payload['folder'] = self.get_folder_by_name_or_moref_path(
                    machine_section['folder']
                )[0]['moref']
                # networking
                spec_payload['networks'] = [
                    {
                        'network': self.get_network_by_name_or_moref(
                            n['network']
                        )[0]['moref'],
                        'type': n['type'],
                    }
                    for n in networking_section['interfaces']
                ]
                spec_payload['disks'] = [
                    {
                        "capacity_gb": d['capacity_gb'],
                        "backing_mode": self.get_vm_disk_backing_mode_by_name(
                            d['backing_mode']
                        )[0]['type']
                        if d.get('backing_mode')
                        else 'persistent',
                        "backing_sharing": self.get_vm_disk_backing_sharing_by_name(  # NOQA:
                            d['backing_sharing']
                        )[
                            0
                        ][
                            'type'
                        ]
                        if d.get('backing_sharing')
                        else 'sharingnone',
                    }
                    for d in machine_section['disks']
                ]
                # other
                machine_section['power_on'] = machine_section.get(
                    'power_on', False
                )
                # metadata section
                spec_payload.update(metadata_section)
                spec_payload['built'] = built
                spec_payload['client'] = metadata_section['client']
                # optional
                if metadata_section.get('inform') is not None:
                    spec_payload['inform'] = [
                        validate_email(None, 'inform', i)
                        for i in metadata_section['inform']
                    ]
                if metadata_section.get('vss_service') is not None:
                    service = self.get_vss_service_by_name_label_or_id(
                        metadata_section['vss_service']
                    )[0]['id']
                    spec_payload['vss_service'] = service
                if metadata_section.get('metadata_section') is not None:
                    admin_name = metadata_section['admin']['name']
                    admin_email = metadata_section['admin']['email']
                    admin_phone = metadata_section['admin']['phone']
                    if admin_name and admin_email and admin_phone:
                        validate_email(None, '', admin_email)
                        validate_phone_number(None, '', admin_phone)
                    spec_payload['admin'] = (
                        f"{admin_name}:" f"{admin_phone}:{admin_email}"
                    )
            return spec_payload
        except KeyError as ex:
            raise click.BadParameter(f'Invalid CLI specification: {ex}')

    def yaml(self) -> YAML:
        """Create default yaml parser."""
        if self:
            return yaml.yaml()

    def yaml_load(self, source: str) -> Any:
        """Load YAML from source."""
        return self.yaml().load(source)

    def yaml_dump(self, source: Any) -> str:
        """Dump dictionary to YAML string."""
        return cast(str, yaml.dump_yaml(self.yaml(), source))

    def yaml_dump_stream(
        self, data: Any, stream: Any = None, **kw: Any
    ) -> Optional[str]:
        """Dump yaml to stream."""
        return yaml.dump_yaml(self.yaml(), data, stream, **kw)

    def download_inventory_file(
        self, request_id: int, directory: str = ''
    ) -> Dict:
        """Download inventory file to a given directory."""
        with self.spinner(disable=self.debug):
            file_path = self.download_inventory_result(
                request_id=request_id, directory=directory
            )
        obj = {'file': file_path}

        self.echo(
            format_output(self, [obj], columns=[('FILE', 'file')], single=True)
        )
        return obj

    def wait_for_requests_to(
        self,
        obj,
        required: List[str] = (
            RequestStatus.PROCESSED.name,
            RequestStatus.SCHEDULED.name,
        ),
        wait: int = 5,
        max_tries: int = 720,
        in_multiple: bool = False,
    ):
        """Wait for multiple requests to complete."""
        if not in_multiple:
            objs = [
                dict(
                    _links=dict(request=r_url),
                    status=obj['status'],
                    request=dict(id=os.path.basename(r_url)),
                )
                for r_url in obj['_links']['request']
            ]
        else:
            objs = obj
        wq = WorkerQueue(max_workers=len(objs))

        with wq.join(debug=self.debug):
            for obj in objs:
                wq.put(
                    functools.partial(
                        self.wait_for_request_to,
                        obj=obj,
                        required=required,
                        wait=wait,
                        max_tries=max_tries,
                        threaded=True,
                    )
                )
                wq.spawn_worker()

    def wait_for_request_to(
        self,
        obj: dict,
        required: List[str] = (
            RequestStatus.PROCESSED.name,
            RequestStatus.SCHEDULED.name,
        ),
        wait: int = 5,
        max_tries: int = 720,
        threaded: bool = False,
    ) -> Optional[bool]:
        """Wait for request to given status."""
        # wait
        request_message = {}
        err_status = [
            RequestStatus.ERROR.name,
            RequestStatus.ERROR_RETRY.name,
            RequestStatus.CANCELLED.name,
        ]
        wait_status = [
            RequestStatus.PENDING.name,
            RequestStatus.IN_PROGRESS.name,
            RequestStatus.APPROVAL_REQUIRED.name,
        ]
        _LOGGING.debug(
            f'max tries={max_tries}, wait={wait}, '
            f'required status={",".join(required)}'
        )
        request_id = obj["request"]["id"]
        self.secho(
            f'{EMOJI_UNICODE.get(":hourglass_not_done:")} '
            f'Waiting for request {request_id} to complete... ',
            fg='green',
            nl=True,
        )
        # check for request status
        if 199 < obj['status'] < 300:
            pass
        else:
            raise VssCliError('Invalid response from the API.')
        with self.spinner(disable=self.debug or threaded):
            r_url = obj['_links']['request']
            tries = 0
            while True:
                request = self.request(r_url)
                if 'data' in request:
                    status = request['data']['status']
                    request_message = request['data']['message']
                    if status in required:
                        request_status = True
                        break
                    if status in err_status:
                        request_status = False
                        break
                    elif status in wait_status:
                        pass
                else:
                    request_status = False
                    break
                if tries >= max_tries:
                    raise VssCliError(
                        f'Wait for request timed out after '
                        f'{max_tries * wait} seconds.'
                    )
                tries += 1
                sleep(wait)
        # check result status
        request_message_str = format_output(
            self,
            [request_message],
            columns=const.COLUMNS_REQUEST_WAIT,
            single=True,
        )
        sys.stdout.flush()
        if request_status:
            self.secho(
                f'{EMOJI_UNICODE.get(":party_popper:")} '
                f'Request {request_id} completed successfully:',
                fg='green',
            )
            self.echo(f'{request_message_str}')
        else:
            self.secho(
                f'{EMOJI_UNICODE.get(":worried_face:")} '
                f'Request {request_id} completed with errors:',
                err=True,
                fg='red',
            )
            self.echo(f'{request_message_str}')
            return False
        return True



"""Validators module for the VSS-CLI."""
from datetime import datetime
import json
import logging
import re
from uuid import UUID

import click

from vss_cli import const

_LOGGING = logging.getLogger(__name__)


def validate_phone_number(ctx, param, phone):
    """Validate phone number."""
    if phone:
        phone_regex = (
            r'(\d{3}[-\.\s]??\d{3}[-\.\s]??\d{4}|'
            r'\(\d{3}\)\s*\d{3}[-\.\s]??\d{4}|\d'
            r'{3}[-\.\s]??\d{4})+'
        )
        if not re.match(phone_regex, phone):
            raise click.BadParameter(
                'Value must be in the ' 'following format 416-166-6666'
            )
    return phone


def validate_email(ctx, param, email):
    """Validate email callback method."""
    email_regex = r'([a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+' r'\.[a-zA-Z0-9-.]+$)'
    if not re.match(email_regex, email):
        raise click.BadParameter(
            f'Value must be in the following format user@utoronto.ca: {email}'
        )
    return email


def flexible_email_args(ctx, param, value):
    """Validate multiple emails."""
    emails = []
    for val in value:
        if ',' in val:
            n_val = [i.strip() for i in val.split(',')]
            emails.extend(n_val)
        else:
            emails.append(val)
    for e in emails:
        validate_email(ctx, param, e)
    return emails


def validate_json_type(ctx, param, value):
    """Validate json callback method."""
    try:
        if value is not None:
            return json.loads(value)
    except ValueError as ex:
        _LOGGING.error(f'{ex}')
        raise click.BadParameter(
            f'{param.name} should be a JSON parameter input.'
        )


def validate_admin(ctx, param, value):
    """Validate admin callback method."""
    if value:
        _value = value.split(':')
        if not value or len(_value) < 2:
            raise click.BadParameter(
                'Admin should be in the '
                'following format: '
                'FullName:Phone:email'
            )
        validate_email(ctx, param, _value[2])
        validate_phone_number(ctx, param, _value[1])
        return value


def validate_inform(ctx, param, value):
    """Validate inform callback method."""
    if value:
        _value = value.split(',')
        if not value:
            raise click.BadParameter(
                'Informational contacts format:'
                'email@utoronto.ca,email2@utoronto.ca'
            )
        for email in _value:
            validate_email(ctx, param, email)
        return value


def validate_uuid(ctx, param, value):
    """Validate uuid callback method."""
    if value:
        try:
            _ = UUID(value)
            return value
        except ValueError:
            return False


def validate_vm_moref(ctx, param, value):
    """Validate moref callback method."""
    if value:
        if re.match(r'vm-[0-9]+', value):
            return value
    return False


def retirement_value(ctx, param, value):
    """Process retirement value."""
    if value:
        tup = value.split(',')
        if len(tup) > 1:
            try:
                hours, days, months = value.split(',')
                hours = int(hours)
                days = int(days)
                months = int(months)
            except ValueError:
                raise click.BadParameter(
                    'Value format does not match: <hours>,<days>,<months>'
                )
            return hours, days, months
        else:
            try:
                _ = datetime.strptime(value, const.DEFAULT_DATETIME_FMT)
            except ValueError:
                raise click.BadParameter(
                    f"Value format does not match "
                    f"'{const.DEFAULT_DATETIME_FMT}'"
                )
            return value
    return value



"""Request plugins module for the VSS-CLI."""
from vss_cli.plugins.request_plugins import (  # pylint: disable=unused-import
    change, export, folder, image, inventory, new, retirement, snapshot, vmdk)



"""Compute plugin for VSS CLI (vss-cli)."""
import click
from click_plugins import with_plugins
from pkg_resources import iter_entry_points

from vss_cli.cli import pass_context
from vss_cli.config import Configuration


@with_plugins(iter_entry_points('vss_cli.contrib.compute'))
@click.group('compute', short_help='Manage VMs, networks, folders, etc.')
@pass_context
def cli(ctx: Configuration):
    """Compute related resources.

    Such as virtual machines, networks supported operating systems,
    logical folders, OVA/OVF images, floppy images, ISO images and more.
    """
    with ctx.spinner(disable=ctx.debug) as spinner_cls:
        ctx.load_config(spinner_cls=spinner_cls)


from vss_cli.plugins.compute_plugins import (  # isort:skip
    domain,
    floppy,
    folder,
    image,
    inventory,
    iso,
    net,
    os as compute_os,
    template,
    vm,
)  # pylint: disable=unused-import



from dateutil import parser as dateutil_parser
import requests

from vss_cli.utils.emoji import EMOJI_UNICODE
from vss_cli.vssconst import STATUS_PAGE_ID, STATUS_PAGE_SERVICE

components_url = (
    'https://{page_id}.statuspage.io/'
    'api/v2/components.json'.format(page_id=STATUS_PAGE_ID)
)

upcoming_maints_url = (
    'https://{page_id}.statuspage.io/'
    'api/v2/scheduled-maintenances/'
    'upcoming.json'.format(page_id=STATUS_PAGE_ID)
)


def format_iso8601(s):
    d = dateutil_parser.parse(s)
    return d.strftime("%Y-%m-%d %a %H:%M:%S %z")


def get_component():
    r = requests.get(components_url)
    components = r.json()
    for cmp in components['components']:
        if STATUS_PAGE_SERVICE in cmp['name']:
            icon = EMOJI_UNICODE.get(':question_mark:')
            status = cmp.get('status')
            if status in ['operational']:
                icon = EMOJI_UNICODE.get(':white_heavy_check_mark:')
            elif 'degraded' in status or 'partial' in status:
                icon = EMOJI_UNICODE.get(':warning:')
            elif 'disruption' in status:
                icon = EMOJI_UNICODE.get(':cross_mark:')
            return {
                'created_at': format_iso8601(cmp.get('created_at')),
                'updated_at': format_iso8601(cmp.get('updated_at')),
                'name': cmp.get('name'),
                'description': cmp.get('description'),
                'status': status,
                'icon': icon,
            }
    return None


def get_upcoming_maintenance_by_service():
    r = requests.get(upcoming_maints_url)
    maints = r.json()
    maintenances = list()
    for maint in maints['scheduled_maintenances']:
        for cmp in maint['components']:
            if STATUS_PAGE_SERVICE in cmp['name']:
                updates = maint.get('incident_updates')
                created_at = format_iso8601(maint.get('created_at'))
                scheduled_for = format_iso8601(maint.get('scheduled_for'))
                scheduled_until = format_iso8601(maint.get('scheduled_until'))
                maint_dict = {
                    'name': maint.get('name'),
                    'impact': maint.get('impact'),
                    'link': maint.get('shortlink'),
                    'status': maint.get('status'),
                    'created_at': created_at,
                    'scheduled_for': scheduled_for,
                    'scheduled_until': scheduled_until,
                    'description': updates[0].get('body') if updates else None,
                }
                maintenances.append(maint_dict)
    return maintenances


def check_status():
    # Component status
    component = get_component()
    # Component upcoming maintenance
    upcoming_maints = get_upcoming_maintenance_by_service()
    component['upcoming_maintenances'] = upcoming_maints
    return component



"""Details for the auto-completion."""
import os
from typing import Callable, Dict, List, Tuple  # NOQA

from requests.exceptions import HTTPError

from vss_cli import const
from vss_cli.config import Configuration
from vss_cli.exceptions import VssError


def _init_ctx(ctx: Configuration) -> None:
    """Initialize ctx."""
    # ctx is incomplete thus need to 'hack' around it
    # see bug https://github.com/pallets/click/issues/942
    ctx.client = Configuration(tk=os.environ.get('VSS_TOKEN'))
    ctx.client.endpoint = os.environ.get('VSS_ENDPOINT', None)
    ctx.client.username = os.environ.get('VSS_USER', None)
    ctx.client.password = os.environ.get('VSS_USER_PASS', None)
    ctx.client.timeout = int(
        os.environ.get('VSS_TIMEOUT', str(const.DEFAULT_TIMEOUT))
    )
    ctx.client.config_path = os.environ.get('VSS_CONFIG', const.DEFAULT_CONFIG)
    # fallback to load configuration
    ctx.client.load_config()


def _autocomplete(
    f: Callable,
    incomplete: str,
    attrs: List[str],
    f_kwargs: Dict = None,
    sort_index: int = 0,
    complete_index: int = 0,
) -> List[Tuple[str, str]]:
    """Autocomplete main function."""
    try:
        response = f(**f_kwargs)
    except (HTTPError, VssError):
        response = []

    completions = []  # type: List[Tuple[str, str]]
    if response:
        for obj in response:
            sub_attrs = []
            # start iterating in second item (index 1)
            for attr in attrs[1:]:
                sub_attrs.append(str(obj.get(attr, 'N/A')))
            # first item (index 0) is always the key
            r = (str(obj[attrs[0]]), ' - '.join(sub_attrs))
            completions.append(r)

        completions.sort(key=lambda x: x[sort_index])

        return [c for c in completions if incomplete in c[complete_index]]
    return completions


def table_formats(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Provide table Formats."""
    _init_ctx(ctx)

    completions = [
        ("plain", "Plain tables, no pseudo-graphics to draw lines"),
        ("simple", "Simple table with --- as header/footer (default)"),
        ("github", "Github flavored Markdown table"),
        ("grid", "Formatted as Emacs 'table.el' package"),
        ("fancy_grid", "Draws a fancy grid using box-drawing characters"),
        ("pipe", "PHP Markdown Extra"),
        ("orgtbl", "org-mode table"),
        ("jira", "Atlassian Jira Markup"),
        ("presto", "Formatted as PrestoDB cli"),
        ("psql", "Formatted as Postgres psql cli"),
        ("rst", "reStructuredText"),
        ("mediawiki", "Media Wiki as used in Wikpedia"),
        ("moinmoin", "MoinMain Wiki"),
        ("youtrack", "Youtrack format"),
        ("html", "HTML Markup"),
        ("pretty", "HTML escaping."),
        ("latex", "LaTeX markup, replacing special characters"),
        ("latex_raw", "LaTeX markup, no replacing of special characters"),
        (
            "latex_booktabs",
            "LaTex markup using spacing and style from `booktabs",
        ),
        ("textile", "Textile"),
        ("tsv", "Tab Separated Values"),
        ("csv", "Comma Separated Values"),
    ]

    completions.sort()

    return [c for c in completions if incomplete in c[0]]


def vm_templates(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM templates."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_templates,
        incomplete,
        ['moref', 'name'],
        complete_index=1,
        sort_index=1,
        f_kwargs={"short": 1, "show_all": True, "per_page": 2000},
    )


def virtual_machines(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VMs."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_vms,
        incomplete,
        ['moref', 'name'],
        complete_index=1,
        sort_index=1,
        f_kwargs={"show_all": True, "short": 1, "per_page": 2000},
    )


def vm_controller_scsi_sharing(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM controller SCSI types."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_scsi_sharing,
        incomplete,
        ['type', 'description'],
        f_kwargs={'only_type': False},
    )


def vm_controller_scsi_types(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM controller SCSI types."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_scsi_controllers,
        incomplete,
        ['type', 'description'],
        f_kwargs={'only_type': False},
    )


def vm_disk_backing_modes(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM Disk Backing Modes."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_disk_backing_modes,
        incomplete,
        ['type', 'description'],
        f_kwargs={'only_type': False},
    )


def vm_disk_sharing(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM Disk Backing Modes."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_disk_sharing,
        incomplete,
        ['type', 'description'],
        f_kwargs={'only_type': False},
    )


def domains(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete Domains."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_domains,
        incomplete,
        ['moref', 'name'],
        complete_index=1,
        sort_index=1,
        f_kwargs={"short": 1, "show_all": True, "per_page": 2000},
    )


def folders(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM folders."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_folders,
        incomplete,
        ['moref', 'path'],
        sort_index=1,
        complete_index=1,
        f_kwargs={
            "show_all": True,
            "sort": "path,desc",
            "per_page": 500,
            "short": 1,
        },
    )


def networks(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete Networks."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_networks,
        incomplete,
        ['moref', 'label'],
        sort_index=1,
        complete_index=1,
        f_kwargs={"show_all": True, "sort": "label,desc", "per_page": 500},
    )


def operating_systems(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM operating systems."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_os,
        incomplete,
        ['guest_id', 'full_name'],
        f_kwargs={"show_all": True, "sort": "guest_id,desc", "per_page": 500},
    )


def vss_services(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VSS Services."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_vss_services,
        incomplete,
        ['id', 'label'],
        sort_index=1,
        complete_index=1,
        f_kwargs={"show_all": True, "sort": "label,desc", "per_page": 500},
    )


def isos(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete ISO images."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_isos,
        incomplete,
        attrs=['id', 'path'],
        sort_index=1,
        complete_index=1,
        f_kwargs={"show_all": True, "sort": "path,desc", "per_page": 500},
    )


def clib_deployable_items(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete deployable items from content library."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_content_library_items,
        incomplete,
        attrs=['id', 'name', 'type'],
        sort_index=1,
        complete_index=1,
        f_kwargs={
            "show_all": True,
            "sort": "path,desc",
            "per_page": 500,
            "filter": "type,in,OVF,VM_TEMPLATE",
        },
    )


def vm_images(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM images."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_images,
        incomplete,
        attrs=['id', 'path'],
        sort_index=1,
        complete_index=1,
        f_kwargs={"show_all": True, "sort": "path,desc", "per_page": 500},
    )


def floppies(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM Floppies."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_floppies,
        incomplete,
        attrs=['id', 'path'],
        sort_index=1,
        complete_index=1,
        f_kwargs={"show_all": True, "sort": "path,desc", "per_page": 500},
    )


def inventory_properties(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM Inventory Properties."""
    _init_ctx(ctx)
    try:
        response = ctx.client.request('/inventory/options')
        if response:
            response = response.get('data')
    except (HTTPError, VssError):
        response = []

    completions = []  # type: List[Tuple[str, str]]
    if response:
        for obj in response:
            completions.append((obj['key'], obj['value']))

        completions.sort()

        return [c for c in completions if incomplete in c[0]]
    return completions


def inventory_requests(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM Inventory Requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_inventory_requests,
        incomplete,
        attrs=['id', 'created_on', 'name'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def new_requests(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete New VM Requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_new_requests,
        incomplete,
        attrs=['id', 'vm_moref', 'vm_name'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def change_requests(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete Change VM Requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_change_requests,
        incomplete,
        attrs=['id', 'vm_moref', 'vm_name', 'attribute'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def export_requests(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete Export VM Requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_export_requests,
        incomplete,
        attrs=['id', 'vm_moref', 'vm_name'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def folder_requests(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete Folder Requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_folder_requests,
        incomplete,
        attrs=['id', 'moref', 'action'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def image_sync_requests(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete Image Sync Requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_image_sync_requests,
        incomplete,
        attrs=['id', 'type'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def vmdk_sync_requests(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete Image Sync Requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_vmdk_sync_requests,
        incomplete,
        attrs=['id', 'created_on'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def snapshot_requests(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM Snapshot Requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_snapshot_requests,
        incomplete,
        attrs=['id', 'vm_moref', 'vm_name'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def account_messages(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete Account Messages."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_user_messages,
        incomplete,
        attrs=['id', 'kind', 'subject'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def virtual_nic_types(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM NIC Types."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_nic_types,
        incomplete,
        attrs=['type', 'description'],
        f_kwargs={"only_type": False},
    )


def virtual_hw_types(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM Virtual Hardware Types."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_vmx_types,
        incomplete,
        attrs=['type', 'description'],
        f_kwargs={"only_type": False},
    )


def vss_options(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VSS Options."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_vss_options,
        incomplete,
        attrs=['option', 'description'],
        f_kwargs={"only_option": False},
    )


def groups(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete User Groups."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_groups,
        incomplete,
        attrs=['id', 'name'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def vm_firmware(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM firmware types."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_firmware_types,
        incomplete,
        attrs=['type', 'description'],
        f_kwargs={"only_type": False},
    )


def vm_snapshots(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM snapshots."""
    _init_ctx(ctx)
    if '--source' in args:
        try:
            s_pos = args.index('--source')
            vm_id = args[s_pos + 1]
            return _autocomplete(
                ctx.client.get_vm_snapshots,
                incomplete,
                attrs=['id', 'name', 'description'],
                f_kwargs={'vm_id': vm_id},
            )
        except IndexError:
            return []
    return []


def retirement_requests(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM retirement requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_retirement_requests,
        incomplete,
        attrs=['id', 'vm_moref', 'vm_name', 'retire_on'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def vm_retirement_requests(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM retirement requests by VM."""
    _init_ctx(ctx)
    if '--source' in args:
        try:
            s_pos = args.index('--source')
            vm_id = args[s_pos + 1]
            return _autocomplete(
                ctx.client.get_vm_retirement_requests,
                incomplete,
                attrs=['id', 'name', 'description'],
                f_kwargs={'vm_id': vm_id},
            )
        except IndexError:
            return []
    return []



"""Compute Callback module for VSS CLI (vss-cli)."""
import json
from pathlib import Path
from typing import Optional, Tuple

from click.exceptions import BadArgumentUsage, BadParameter

from vss_cli.autocompletion import _init_ctx
from vss_cli.config import Configuration
from vss_cli.helper import to_tuples


def process_networks_opt(ctx: Configuration, param, value):
    """Process network option."""
    _init_ctx(ctx)
    if value is not None:
        networks = list()
        for nic in value:
            _nic = to_tuples(nic)[0]
            _network = _nic[0]
            if len(_nic) > 1:
                _type = ctx.client.get_vm_nic_type_by_name(_nic[1])
                _type = _type[0]['type']
            else:
                _type = 'vmxnet3'
            _net = ctx.client.get_network_by_name_or_moref(_network)
            networks.append({'network': _net[0]['moref'], 'type': _type})
        return networks


def process_scsi_opt(ctx: Configuration, param, value):
    """Process SCSI spec option.

    <type>=<sharing>

     or

    {
     "type": "paravirtual",
     "sharing": "nosharing",
     "bus": 0,
    }
    """
    _init_ctx(ctx)
    if value is not None:
        devices = list()
        _sharing = 'nosharing'
        for dev in value:
            scsi = {}
            try:
                _dev = json.loads(dev)
                if not isinstance(_dev, dict):
                    raise BadParameter(
                        'Must be an object with at least the type attribute'
                    )
                is_json = True
            except ValueError:
                is_json = False
                _dev = to_tuples(dev)[0]
            # get values either from tuple or dict
            try:
                if is_json:
                    _type = _dev['type']
                else:
                    _type = _dev[0]
                # lookup device type
                _type = ctx.client.get_vm_scsi_type_by_name(_type)
                _type = _type[0]['type']
                scsi['type'] = _type
            except KeyError:
                raise BadParameter('at least type must be provided')

            if is_json:
                if _dev.get('sharing') is not None:
                    _sharing = ctx.client.get_vm_scsi_sharing_by_name(
                        _dev.get('sharing')
                    )
                    _sharing = _sharing[0]['type']
                    scsi['sharing'] = _sharing
                if _dev.get('bus') is not None:
                    try:
                        _bus = int(_dev['bus'])
                    except ValueError:
                        raise BadParameter('bus must be a number')
                    scsi['bus'] = _bus
            else:
                if len(_dev) > 1:
                    _sharing = ctx.client.get_vm_scsi_sharing_by_name(_dev[1])
                    _sharing = _sharing[0]['type']
                    scsi['sharing'] = _sharing
                else:
                    scsi['sharing'] = _sharing
            devices.append(scsi)
        return devices


def process_disk_opt(ctx: Configuration, param, value):
    """Process Disk spec option.

    <capacity>=<backing_mode>=<backing_sharing>=<backing_vmdk>

     or

    {
     "capacity_gb": 100,
     "backing_mode": "persistent",
     "backing_sharing": "sharingnone",
     "backing_vmdk": "[vssUser-xfers] vskey/user/disk-0.vmdk",
     "scsi": 1
    }
    """
    _init_ctx(ctx)
    if value is not None:
        devices = list()
        _backing_mode = 'persistent'
        _sharing_mode = 'sharingnone'
        for dev in value:
            disk = {}
            try:
                _dev = json.loads(dev)
                if not isinstance(_dev, dict):
                    raise ValueError(
                        'Must be an object with at least capacity_gb'
                    )
                is_json = True
            except ValueError:
                is_json = False
                _dev = to_tuples(dev)[0]

            try:
                if is_json:
                    _capacity = int(_dev['capacity_gb'])
                else:
                    _capacity = int(_dev[0])
                disk['capacity_gb'] = _capacity
            except ValueError:
                raise BadParameter('capacity must be a number')
            if is_json:
                if _dev.get('backing_mode'):
                    _backing_mode = ctx.client.get_vm_disk_backing_mode_by_name(  # NOQA:
                        _dev.get('backing_mode')
                    )
                    _backing_mode = _backing_mode[0]['type']
                    disk['backing_mode'] = _backing_mode
                if _dev.get('backing_sharing'):
                    _sharing_mode = ctx.client.get_vm_disk_backing_sharing_by_name(  # NOQA:
                        _dev.get('backing_sharing')
                    )
                    _sharing_mode = _sharing_mode[0]['type']
                    disk['backing_sharing'] = _sharing_mode
                if _dev.get('backing_vmdk'):
                    _backing_vmdk = ctx.client.get_vmdk_by_name_path_or_id(
                        _dev.get('backing_vmdk')
                    )
                    disk['backing_vmdk'] = _backing_vmdk[0]['path']
                if _dev.get('scsi') is not None:
                    try:
                        _scsi_crllr = int(_dev.get('scsi'))
                    except ValueError:
                        raise BadParameter('scsi must be a number')
                    disk['scsi'] = _scsi_crllr
            else:
                if len(_dev) > 1:
                    _spec = to_tuples(_dev[1])[0]
                    if len(_spec) > 0:
                        _backing_mode = ctx.client.get_vm_disk_backing_mode_by_name(  # NOQA:
                            _spec[0]
                        )
                        _backing_mode = _backing_mode[0]['type']
                        disk['backing_mode'] = _backing_mode
                    if len(_spec) > 1:
                        _spec = to_tuples(_spec[1])[0]
                        _sharing_mode = ctx.client.get_vm_disk_backing_sharing_by_name(  # NOQA
                            _spec[0]
                        )
                        _sharing_mode = _sharing_mode[0]['type']
                        disk['backing_sharing'] = _sharing_mode
                    if len(_spec) > 1:
                        _backing_vmdk = ctx.client.get_vmdk_by_name_path_or_id(
                            _spec[1]
                        )
                        _backing_vmdk = _backing_vmdk[0]['path']
                        disk['backing_vmdk'] = _backing_vmdk
            devices.append(disk)
        return devices


def process_options(ctx: Configuration, param, key_value):
    """Process options."""
    _init_ctx(ctx)
    try:
        _options = to_tuples(','.join(key_value))
        options = [{opt[0]: opt[1]} for opt in _options]
        return options
    except Exception:
        raise BadArgumentUsage(f'{param} must be key=value strings')


def process_firmware(ctx: Configuration, param, value):
    """Process firmware options."""
    _init_ctx(ctx)
    try:
        _firmware = ctx.client.get_vm_firmware_by_type_or_desc(value)
        return _firmware[0]['type']
    except Exception:
        valid = ctx.client.get_supported_firmware_types()
        raise BadArgumentUsage(f'{param} must be one of: {", ".join(valid)}')


def process_user_data(
    ctx: Configuration, param, value
) -> Optional[Tuple[str, str]]:
    """Process user_data."""
    from pyvss.helper import compress_encode_string

    _init_ctx(ctx)
    if value:
        try:
            fp = Path(value)
            txt = fp.read_text()
            return (
                compress_encode_string(txt),
                'gzip+base64',
            )
        except Exception:
            raise BadArgumentUsage(f'{param} must a valid file path.')



"""Helper module for the vss-cli compute plugins."""
from typing import Dict, Optional, Tuple, Union

import click


def process_retirement_new(
    retire_type: str,
    retire_value: Union[Tuple[int, int, int], str],
    retire_warning: Optional[int] = None,
) -> Dict:
    """Process retirement for new vm commands."""
    if not all([retire_type, retire_value]):
        raise click.BadParameter(
            'Retirement settings require at least: '
            '--retire-type and --retire-value'
        )
    if retire_type == 'timedelta':
        retire = {
            'value': {
                'type': retire_type,
                'hours': retire_value[0],
                'days': retire_value[1],
                'months': retire_value[2],
            }
        }
    else:
        retire = {'value': {'type': retire_type, 'datetime': retire_value}}
    if retire_warning:
        retire['warning'] = {'days': retire_warning}
    else:
        confirmation = click.confirm(
            'No warning will be sent for confirmation or cancellation. \n'
            'Retirement request will proceed when specified. \n'
            'Are you sure?'
        )
        if not confirmation:
            raise click.ClickException('Cancelled by user.')
    return retire



"""The exceptions used by VSS CLI."""
from pyvss.exceptions import VssError


class VssCliError(Exception):
    """General VSS CLI exception occurred."""


class ValidationError(Exception):
    """Validation error exception"""



"""SSH Key Management plugin for VSS CLI (vss-cli)."""
import logging
import os

import click

from vss_cli import const, rel_opts as so
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.exceptions import VssCliError
from vss_cli.helper import format_output

_LOGGING = logging.getLogger(__name__)


@click.group('key', short_help='Manage your SSH Public Keys.')
@pass_context
def cli(ctx: Configuration):
    """Manage your SSH Public Keys."""
    with ctx.spinner(disable=ctx.debug):
        ctx.load_config()


@cli.command('ls', short_help='List user SSH Public Keys.')
@so.filter_opt
@so.sort_opt
@so.all_opt
@so.count_opt
@so.page_opt
@pass_context
def key_ls(ctx: Configuration, filter_by, page, sort, show_all, count):
    """List ssh public keys.

    Filter list in the following format <field_name>=<operator>,<value>
    where operator is eq, ne, lt, le, gt, ge, like, in.
    For example: type eq,ssh-rsa

    vss-cli key ls -f type=eq,ssh-rsa

    Sort list in the following format <field_name>=<asc|desc>. For example:

    vss-cli key ls -s created_on=desc
    """
    columns = ctx.columns or const.COLUMNS_SSH_KEY_MIN
    params = dict()
    if all(filter_by):
        params['filter'] = ';'.join(filter_by)
    if all(sort):
        params['sort'] = ';'.join(sort)
    # make request
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_ssh_keys(
            show_all=show_all, per_page=count, **params
        )
    # format output
    output = format_output(ctx, obj, columns=columns)
    # page results
    if page:
        click.echo_via_pager(output)
    else:
        ctx.echo(output)


@cli.command('get', help='Display user key info.')
@click.argument('kid', type=int, required=True)
@pass_context
def key_get(ctx: Configuration, kid):
    """Get ssh key info."""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_ssh_key(kid)
    columns = ctx.columns or const.COLUMNS_SSH_KEY
    # format output
    ctx.echo(format_output(ctx, [obj], columns=columns, single=True))


@cli.command('mk', help='Create SSH Public Key.')
@click.argument('path_or_key', type=click.STRING, required=True, nargs=1)
@pass_context
def key_mk(ctx, path_or_key):
    """Create ssh-key record."""
    with ctx.spinner(disable=ctx.debug):
        if os.path.isfile(path_or_key):
            obj = ctx.create_user_ssh_key_path(path_or_key)
        else:
            obj = ctx.create_user_ssh_key(path_or_key)
    # defining columns
    columns = ctx.columns or const.COLUMNS_SSH_KEY_MIN
    # if key has been created print
    if obj:
        # format output
        ctx.echo(format_output(ctx, [obj], columns=columns, single=True))
    else:
        raise VssCliError(
            'Cloud not create key. '
            'Verify if a valid SSH Public Key'
            ' has been provided.'
        )


@cli.command('rm', help='Delete SSH Public Key.')
@click.argument('kid', type=int, required=True, nargs=-1)
@click.option('-s', '--summary', is_flag=True, help='Print request summary')
@pass_context
def key_rm(ctx: Configuration, kid, summary):
    """Delete ssh keys."""
    result = []
    with click.progressbar(kid) as ids:
        for i in ids:
            result.append(ctx.delete_user_ssh_key(i))
    if summary:
        for res in result:
            ctx.echo(
                format_output(
                    ctx, [res], columns=[('STATUS', 'status')], single=True
                )
            )



"""Upgrade plugin for VSS CLI (vss-cli)."""
import logging
from subprocess import call
import sys

import click

from vss_cli import vssconst
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.utils.emoji import EMOJI_UNICODE

_LOGGING = logging.getLogger(__name__)


@click.group(
    'upgrade',
    invoke_without_command=True,
    short_help='Upgrade VSS CLI and dependencies.',
)
@click.argument(
    'upstream',
    type=click.Choice(['stable', 'develop', 'branch']),
    default='stable',
)
@click.option(
    '--git-branch', '-b', help='GitLab repository branch', required=False
)
@pass_context
def cli(ctx: Configuration, upstream, git_branch):
    """Upgrade existing install of VSS CLI to the latest version.

    For example, to upgrade to the stable version:

    vss-cli upgrade stable

    To upgrade VSS CLI to the latest develop build:

    vss-cli upgrade develop

    To upgrade to a given official vss-cli GitLab repository:

    vss-cli upgrade --git-branch=issue-145 branch
    """
    lookup = {
        'stable': {'pkg': 'vss-cli', 'args': ['--upgrade']},
        'develop': {'pkg': 'vss-cli', 'args': ['--upgrade', '--pre']},
        'branch': {
            'pkg': f'https://gitlab-ee.eis.utoronto.ca'
            f'/vss/vss-cli/-/archive/'
            f'{git_branch}/vss-cli-{git_branch}.zip',
            'args': ['--upgrade'],
        },
    }
    try:
        cmd_lookup = lookup[upstream]
        if upstream == 'branch':
            if not git_branch:
                raise click.BadOptionUsage(
                    'branch',
                    'Missing --branch/-b option. ' 'Try master or develop.',
                )
    except ValueError as ex:
        _LOGGING.error(f'Invalid upstream {upstream}: {ex}')
        raise click.BadArgumentUsage(f'Invalid upstream {upstream}')
    # assemble command
    cmd_args_str = ' '.join(cmd_lookup['args'])
    cmd_bin = sys.executable
    # execute command
    cmd_str = f"{cmd_bin} -m pip install {cmd_args_str} {cmd_lookup['pkg']}"
    _LOGGING.debug(f'Executing {cmd_str}')
    # calling command
    exit_code = call(cmd_str.split())
    if exit_code > 0:
        raise click.ClickException(
            f'Could not perform upgrade, please try: ' f'\n\t{cmd_str}'
        )
    else:
        _LOGGING.debug(f'Successfully executed upgrade command: {cmd_str}')
    # all done
    ctx.secho(
        f'Successfully executed upgrade command '
        f'{EMOJI_UNICODE.get(":white_heavy_check_mark:")}',
        fg='green',
    )



import contextlib
import itertools
import logging
import queue
import sys

import click
import click_spinner
import click_threading

_LOGGING = logging.getLogger(__name__)


class WorkerQueue(object):
    '''
    Based on https://github.com/pimutils/vdirsyncer
    A simple worker-queue setup.
    Note that workers quit if queue is empty. That means you have to first put
    things into the queue before spawning the worker!
    '''

    def __init__(self, max_workers):
        self._queue = queue.Queue()
        self._workers = []
        self._max_workers = max_workers
        self._shutdown_handlers = []

        # According to http://stackoverflow.com/a/27062830, those are
        # thread safe compared to increasing a simple integer variable.
        self.num_done_tasks = itertools.count()
        self.num_failed_tasks = itertools.count()

    def shutdown(self):
        while self._shutdown_handlers:
            try:
                self._shutdown_handlers.pop()()
            except Exception:
                pass

    def _worker(self):
        while True:
            try:
                func = self._queue.get(False)
            except queue.Empty:
                break

            try:
                func()
            except Exception as ex:
                _LOGGING.warning(ex)
                next(self.num_failed_tasks)
            finally:
                self._queue.task_done()
                next(self.num_done_tasks)
                if not self._queue.unfinished_tasks:
                    self.shutdown()

    def spawn_worker(self):
        if self._max_workers and len(self._workers) >= self._max_workers:
            return

        t = click_threading.Thread(target=self._worker)
        t.start()
        self._workers.append(t)

    @contextlib.contextmanager
    def join(self, debug=False):
        assert self._workers or not self._queue.unfinished_tasks
        ui_worker = click_threading.UiWorker()
        self._shutdown_handlers.append(ui_worker.shutdown)
        _echo = click.echo
        _secho = click.secho

        with ui_worker.patch_click():
            yield

            if not self._workers:
                # Ugly hack, needed because
                # ui_worker is not running.
                click.echo = _echo
                click.secho = _secho
                _LOGGING.critical('Nothing to do.')
                sys.exit(5)
            with click_spinner.spinner(disable=debug):
                ui_worker.run()
                self._queue.join()
                for worker in self._workers:
                    worker.join()

        tasks_failed = next(self.num_failed_tasks)
        tasks_done = next(self.num_done_tasks)

        if tasks_failed > 0:
            _LOGGING.error(
                '{} out of {} tasks failed.'.format(tasks_failed, tasks_done)
            )
            sys.exit(1)

    def put(self, f):
        return self._queue.put(f)



"""VSS CLI (vss-cli)."""
import logging
import os
import platform
import sys
from typing import List, Optional, Union, cast

import click
from click.core import Command, Context, Group
import click_log
from pyvss.const import __version__ as pyvss_version

import vss_cli.autocompletion as autocompletion
from vss_cli.config import Configuration
import vss_cli.const as const
from vss_cli.helper import to_tuples

click_log.basic_config()

_LOGGER = logging.getLogger(__name__)

CONTEXT_SETTINGS = dict(auto_envvar_prefix='VSS')

pass_context = click.make_pass_decorator(  # pylint: disable=invalid-name
    Configuration, ensure=True
)


def run() -> None:
    """Run entry point.

    Wraps click for full control over exception handling in Click.
    """
    # A hack to see if exception details should be printed.
    exceptionflags = ['-x']
    verbose = [c for c in exceptionflags if c in sys.argv]

    try:
        # Could use cli.invoke here to use the just created context
        # but then shell completion will not work. Thus calling
        # standalone mode to keep that working.
        result = cli.main(standalone_mode=False)
        if isinstance(result, int):
            sys.exit(result)

    # Exception handling below is done to use logger
    # and mimic as close as possible what click would
    # do normally in its main()
    except click.ClickException as ex:
        ex.show()  # let Click handle its own errors
        sys.exit(ex.exit_code)
    except click.Abort:
        _LOGGER.critical("Aborted!")
        sys.exit(1)
    except Exception as ex:  # pylint: disable=broad-except
        if verbose:
            _LOGGER.exception(ex)
        else:
            _LOGGER.error("%s: %s", type(ex).__name__, ex)
            _LOGGER.info(
                "Run with %s to see full exception information.",
                " or ".join(exceptionflags),
            )
        sys.exit(1)


class VssCli(click.Group):
    """The ITS Private Cloud Command-line."""

    def list_commands(self, ctx: Context) -> List[str]:
        """List all command available as plugin."""
        cmd_folder = os.path.abspath(
            os.path.join(os.path.dirname(__file__), 'plugins')
        )

        commands = []
        for filename in os.listdir(cmd_folder):
            if filename.endswith('.py') and not filename.startswith('__'):
                commands.append(filename[:-3])
        commands.sort()
        _LOGGER.debug(f'Loading {commands}')
        return commands

    def get_command(
        self, ctx: Context, cmd_name: str
    ) -> Optional[Union[Group, Command]]:
        """Import the commands of the plugins."""
        try:
            mod = __import__(
                f'{const.PACKAGE_NAME}.plugins.{cmd_name}', {}, {}, ['cli']
            )
            _LOGGER.debug(f'Loading {mod}')
        except ImportError as ex:
            _LOGGER.debug(
                f'Error loading plugin'
                f' {cmd_name} {type(ex).__name__}: {ex}'
            )
            return None
        return cast(Union[Group, Command], mod.cli)


def _default_token() -> Optional[str]:
    return os.environ.get('VSS_TOKEN', os.environ.get('VSS_API_TOKEN', None))


@click.command(cls=VssCli, context_settings=CONTEXT_SETTINGS)
@click.option(
    '--endpoint',
    '-e',
    help='The Cloud API endpoint URL',
    default=None,
    show_default=True,
    envvar='VSS_ENDPOINT',
)
@click.option(
    '--config',
    '-c',
    default=const.DEFAULT_CONFIG,
    help='Configuration file',
    envvar='VSS_CONFIG',
)
@click.option(
    '--token',
    '-t',
    default=_default_token,
    help='The Bearer token for the VSS API.',
    envvar='VSS_TOKEN',
)
@click.option(
    '--username',
    '-u',
    default=None,  # type: ignore
    help='The API username for VSS API.',
    envvar='VSS_USER',
)
@click.option(
    '--password',
    '-p',
    default=None,  # type: ignore
    help='The API password for VSS API.',
    envvar='VSS_USER_PASS',
)
@click.option(
    '--totp',
    default=_default_token,
    help='Timed One Time Password.',
    envvar='VSS_API_USER_OTP',
)
@click.option(
    '--timeout',
    help='Timeout for network operations.',
    type=click.INT,
    envvar='VSS_TIMEOUT',
    show_default=True,
)
@click_log.simple_verbosity_option(logging.getLogger(), "--loglevel", "-l")
@click.option(
    '-v',
    '--verbose',
    is_flag=True,
    default=False,
    envvar='VSS_VERBOSE',
    help='Enables verbose mode.',
)
@click.option(
    '--debug',
    is_flag=True,
    default=False,
    envvar='VSS_DEBUG',
    help='Enables debug mode.',
)
@click.option(
    '-x',
    'showexceptions',
    is_flag=True,
    default=False,
    envvar='VSS_EXCEPTIONS',
    help="Print back traces when exception occurs.",
)
@click.option(
    '-o',
    '--output',
    help="Output format (default: auto).",
    type=click.Choice(['json', 'yaml', 'table', 'auto', 'ndjson']),
    default=None,
    envvar='VSS_OUTPUT',
    show_default=True,
)
@click.option(
    '--table-format',
    default=None,
    envvar='VSS_TABLE',
    help="Which table format to use (default: simple)",
    shell_complete=autocompletion.table_formats,
)
@click.option(
    '--columns',
    default=None,
    help=(
        'Custom columns key=value list.'
        ' Example: VM=moref,PROVISIONED=storage.provisionedGB'
    ),
)
@click.option(
    '--columns-width',
    default=None,
    type=click.INT,
    envvar='VSS_COL_WIDTH',
    help='Truncates column values (0: auto, -1: disable).',
)
@click.option(
    '--wait/--no-wait',
    is_flag=True,
    help='wait for request(s) to complete',
    default=None,
    envvar='VSS_WAIT_FOR_REQUESTS',
    show_default=True,
)
@click.option(
    '-n',
    '--no-headers',
    default=False,
    is_flag=True,
    help="When printing tables don\'t use headers (default: print headers)",
)
@click.option(
    '-s',
    '--sort-by',
    default=None,
    help='Sort table by the jsonpath expression. Example: updated_on',
)
@click.option(
    '--webdav-server',
    '-w',
    help='The Webdav server.',
    default=None,
    show_default=True,
    envvar='VSS_WEBDAV_SERVER',
)
@click.version_option(
    version=f'{const.__version__} ('
    f'pyvss/{pyvss_version} '
    f'{platform.python_implementation()}/{platform.python_version()})',
    message='%(prog)s v%(version)s',
)
@pass_context
def cli(
    ctx: Configuration,
    verbose: bool,
    endpoint: Optional[str],
    token: Optional[str],
    username: Optional[str],
    password: Optional[str],
    totp: Optional[str],
    config: str,
    output: str,
    timeout: Optional[int],
    debug: bool,
    showexceptions: bool,
    columns: str,
    columns_width: int,
    no_headers: bool,
    table_format: str,
    sort_by: Optional[str],
    wait: Optional[bool],
    webdav_server: Optional[str],
):
    """Command line interface for the ITS Private Cloud."""
    ctx.verbose = verbose
    ctx.endpoint = endpoint
    ctx.token = token
    ctx.config_path = config
    ctx.username = username
    ctx.password = password
    ctx.totp = totp
    ctx.timeout = timeout
    ctx.output = output
    ctx.debug = debug
    ctx.showexceptions = showexceptions
    ctx.columns = to_tuples(columns)
    ctx.columns_width = columns_width
    ctx.no_headers = no_headers
    ctx.table_format = table_format
    ctx.sort_by = sort_by  # type: ignore
    ctx.wait_for_requests = wait
    ctx.webdav_server = webdav_server

    _LOGGER.debug(f"Using settings: {ctx}")



"""Request Management plugin for VSS CLI (vss-cli)."""
import click

from vss_cli.cli import pass_context
from vss_cli.config import Configuration


@click.group('request', short_help='Manage various requests')
@pass_context
def cli(ctx: Configuration):
    """Track request status and details."""
    with ctx.spinner(disable=ctx.debug):
        ctx.load_config()


from vss_cli.plugins.request_plugins import (  # isort:skip
    change,
    export,
    folder,
    image,
    inventory,
    new,
    snapshot,
    vmdk,
)  # pylint: disable=unused-import



"""Status plugin for VSS CLI (vss-cli)."""
import logging

import click

from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.hcio import check_status as check_hcio
from vss_cli.helper import format_output
from vss_cli.sstatus import check_status as check_sstatus

_LOGGING = logging.getLogger(__name__)


@click.group(
    'status', invoke_without_command=True, short_help='Check API Status.'
)
@pass_context
def cli(ctx: Configuration):
    """Check API from https://healthchecks.io."""
    ctx.set_defaults()
    with ctx.spinner(disable=ctx.debug):
        objs = [check_hcio(), check_sstatus()]
    if click.get_current_context().invoked_subcommand is None:
        columns = [('NAME', 'name'), ('STATUS', 'status'), ('ICON', 'icon')]
        columns = ctx.columns or columns
        click.echo(format_output(ctx, objs, columns=columns))


@cli.command('vss', short_help='Check VSS Status.')
@click.option(
    '--maintenance',
    '-m',
    help='Show any scheduled maintenance',
    is_flag=True,
    default=False,
)
@pass_context
def get_vss_status_and_maint(ctx: Configuration, maintenance):
    """Check VSS Status from https://www.systemstatus.utoronto.ca/."""
    ctx.set_defaults()
    with ctx.spinner(disable=ctx.debug):
        system_status = check_sstatus()
    if maintenance:
        columns = [
            ('NAME', 'name'),
            ('IMPACT', 'impact'),
            ('STATUS', 'status'),
            ('DESCRIPTION', 'description[0:100]'),
            ('SCHEDULED', 'scheduled_for'),
        ]
        columns = ctx.columns or columns
        dat = system_status.get('upcoming_maintenances')
        click.echo(format_output(ctx, dat, columns=columns))
    else:
        columns = [
            ('NAME', 'name'),
            ('DESCRIPTION', 'description'),
            ('STATUS', 'status'),
            ('UPDATED', 'updated_at'),
            ('MAINTENANCE', 'upcoming_maintenances[*].name'),
        ]
        columns = ctx.columns or columns
        click.echo(
            format_output(ctx, [system_status], columns=columns, single=True)
        )



"""Compute Shared Arguments plugin for VSS CLI (vss-cli)."""

import click

from vss_cli import autocompletion
from vss_cli.plugins.compute_plugins import callbacks

extra_config_arg = click.argument(
    'key-value',
    type=click.STRING,
    required=True,
    nargs=-1,
    callback=callbacks.process_options,
)
firmware_arg = click.argument(
    'firmware',
    shell_complete=autocompletion.vm_firmware,
    required=True,
    callback=callbacks.process_firmware,
)



#!/usr/bin/env python3
"""Setup script for ITS Private Cloud CLI."""
import codecs
from datetime import datetime as dt
import os
import re
from typing import List

from setuptools import find_packages, setup

# shared consts using approach suggested at
# https://stackoverflow.com/questions/17583443/what-is-the-correct-way-to-share-package-version-with-setup-py-and-the-package


def read(*parts):
    """Read file from current directory."""
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, *parts), 'r') as infofile:
        return infofile.read()


def find_version(*file_paths):
    """Locate version info to share between const.py and setup.py."""
    version_file = read(*file_paths)  # type: ignore
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def load_requirements(requires_file: str = 'requirements.txt') -> List[str]:
    """Load requirements from file"""
    reqs = []
    with open(requires_file, encoding='utf-8') as f:
        _lines = f.read().splitlines()
    for line in _lines:
        line = line.strip('\n')
        if not line.startswith('#') and line != '':
            reqs.append(line)
    return reqs


__VERSION__ = find_version("vss_cli", "const.py")  # type: ignore

REQUIRED_PYTHON_VER = (3, 7, 5)
REQUIRES = load_requirements()
# REQUIRES = [req for req in INSTALL_REQUIRES if req and not re.match(r'[^:]+://', req)]

PACKAGES = find_packages(exclude=['tests', 'tests.*'])
PACKAGE_DATA = {'vss_cli': ['data/*.yaml']}

PROJECT_NAME = 'ITS Private Cloud CLI'
PROJECT_PACKAGE_NAME = 'vss-cli'
PROJECT_LICENSE = 'MIT'
PROJECT_AUTHOR = 'University of Toronto'
PROJECT_COPYRIGHT = f' 2019-{dt.now().year}, {PROJECT_AUTHOR}'
PROJECT_URL = 'https://gitlab-ee.eis.utoronto.ca/vss/vss-cli'
PROJECT_DOCS = 'https://eis.utoronto.ca/~vss/vss-cli'
PROJECT_EMAIL = 'vss-apps@eis.utoronto.ca'
MAINTAINER_EMAIL = 'vss-py@eis.utoronto.ca'

PROJECT_GITLAB_GROUP = 'vss'
PROJECT_GITLAB_REPOSITORY = 'vss-cli'

PYPI_URL = f'https://pypi.python.org/pypi/{PROJECT_PACKAGE_NAME}'
GITLAB_PATH = f'{PROJECT_GITLAB_GROUP}/{PROJECT_GITLAB_REPOSITORY}'
GITLAB_URL = f'https://gitlab-ee.eis.utoronto.ca/{GITLAB_PATH}'

DOWNLOAD_URL = f'{GITLAB_URL}/archive/{__VERSION__}.zip'
PROJECT_URLS = {
    'Bug Reports': f'{GITLAB_URL}/issues',
    'Documentation': f'{PROJECT_DOCS}/',
    'Source': f'{PROJECT_URL}',
}
STOR_REQUIRE = ['webdavclient3==3.14.6']
TESTS_REQUIRE = [
    'flake8==4.0.1',
    'nose==1.3.7',
    'coverage==6.0.2',
    'pytz==2021.3',
    'wheel==0.37.0',  # Otherwise setup.py bdist_wheel does not work
    *STOR_REQUIRE,
]
DEV_REQUIRE = [
    *TESTS_REQUIRE,
    *STOR_REQUIRE,
    'sphinx-rtd-theme==1.0.0',
    'Sphinx==4.2.0',
]

# Allow you to run
# pip install .[test]
# pip install .[dev]
# to get test dependencies included
EXTRAS_REQUIRE = {
    'test': TESTS_REQUIRE,
    'dev': DEV_REQUIRE,
    'stor': STOR_REQUIRE,
}

MIN_PY_VERSION = '.'.join(map(str, REQUIRED_PYTHON_VER))


setup(
    name=PROJECT_PACKAGE_NAME,
    version=__VERSION__,
    url=PROJECT_URL,
    download_url=DOWNLOAD_URL,
    project_urls=PROJECT_URLS,
    author=PROJECT_AUTHOR,
    author_email=PROJECT_EMAIL,
    maintainer_email=MAINTAINER_EMAIL,
    packages=PACKAGES,
    package_data=PACKAGE_DATA,
    license=PROJECT_LICENSE,
    include_package_data=True,
    zip_safe=False,
    install_requires=REQUIRES,
    tests_require=TESTS_REQUIRE,
    extras_require=EXTRAS_REQUIRE,
    python_requires=f'>={MIN_PY_VERSION}',
    entry_points={
        'console_scripts': [
            'vss-cli = vss_cli.cli:run',
            'vss = vss_cli.cli:run',
        ]
    },
)



"""Compute Network plugin for VSS CLI (vss-cli)."""
import logging

import click

from vss_cli import const, rel_opts as so
import vss_cli.autocompletion as autocompletion
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.exceptions import VssCliError
from vss_cli.helper import format_output
from vss_cli.plugins.compute import cli

_LOGGING = logging.getLogger(__name__)


@cli.group('net', short_help='List available virtual networks')
@pass_context
def cli(ctx: Configuration):
    """List available virtual networks."""


@cli.command('ls', short_help='list virtual networks.')
@so.filter_opt
@so.sort_opt
@so.all_opt
@so.page_opt
@pass_context
def network_ls(ctx: Configuration, filter_by, show_all, sort, page):
    """List available virtual networks.

    Filter by path or name name=<name> or moref=<moref>.
    For example:

    vss-cli compute net ls -f name=public
    """
    params = dict(expand=1, sort='name,asc')
    if all(filter_by):
        params['filter'] = ';'.join(filter_by)
    if all(sort):
        params['sort'] = ';'.join(sort)
    # get objects
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_networks(show_all=show_all, **params)
    # set columns
    columns = ctx.columns or const.COLUMNS_NET_MIN
    # format
    output = format_output(ctx, obj, columns=columns)
    # page results
    if page:
        click.echo_via_pager(output)
    else:
        ctx.echo(output)


@cli.group(
    'get', help='Given virtual network info.', invoke_without_command=True
)
@click.argument(
    'name_or_moref',
    type=click.STRING,
    required=True,
    shell_complete=autocompletion.networks,
)
@pass_context
def network_get(ctx: Configuration, name_or_moref):
    """Get Network object information."""
    _net = ctx.get_network_by_name_or_moref(name_or_moref)
    ctx.moref = _net[0]['moref']
    if click.get_current_context().invoked_subcommand is None:
        columns = ctx.columns or const.COLUMNS_NET
        obj = ctx.get_network(ctx.moref)
        ctx.echo(format_output(ctx, [obj], columns=columns, single=True))


@network_get.command('vms', help='List vms on network.')
@click.option(
    '-p', '--page', is_flag=True, help='page results in a less-like format'
)
@pass_context
def net_get_vms(ctx: Configuration, page):
    """List virtual machines using current network."""
    objs = ctx.get_vms_by_network(ctx.moref)
    if not objs:
        raise VssCliError(
            f'Either network {ctx.moref} does not exist, '
            f'or you do not have permission to access.'
        )
    columns = ctx.columns or const.COLUMNS_VM_MIN
    output = format_output(ctx, objs, columns=columns)
    # page
    if page:
        click.echo_via_pager(output)
    else:
        ctx.echo(output)


@network_get.command('perm', help='List network permissions.')
@click.option(
    '-p', '--page', is_flag=True, help='page results in a less-like format'
)
@pass_context
def net_get_permission(ctx: Configuration, page):
    """Obtain network group or user permissions."""
    obj = ctx.get_network_permission(ctx.moref)
    if not obj:
        raise VssCliError(
            f'Either network {ctx.moref} does not exist, '
            f'or you do not have permission to access.'
        )
    columns = ctx.columns or const.COLUMNS_PERMISSION
    output = format_output(ctx, obj, columns=columns)
    # page
    if page:
        click.echo_via_pager(output)
    else:
        ctx.echo(output)



"""Compute Files plugin for VSS CLI (vss-cli)."""
import logging

import click

from vss_cli import const, rel_opts as so
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output
from vss_cli.plugins.compute import cli

_LOGGING = logging.getLogger(__name__)


@cli.group('vmdk', short_help='Manage personal VMDK files.')
@pass_context
def compute_vmdk(ctx):
    """Manage VM images.

    Available files in your personal space.
    """


@compute_vmdk.group('personal', short_help='Browse current vmdks.')
@pass_context
def compute_vmdk_personal(ctx):
    """Available VMDK in your personal VSKEY-STOR space."""
    pass


@compute_vmdk_personal.command('ls', short_help='List personal VMDK Files.')
@click.option(
    '-p', '--page', is_flag=True, help='page results in a less-like format'
)
@pass_context
def compute_vmdk_personal_ls(ctx: Configuration, page):
    """List VMDK files stored in your personal VSKEY-STOR space.

    If the VMDK you uploaded is not listed, use the sync command
    and try again.

    vss-cli compute vmdk personal sync
    vss-cli compute vmdk personal ls
    """
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_vmdks()
    # format
    columns = ctx.columns or const.COLUMNS_IMAGE
    output = format_output(ctx, obj, columns=columns)
    # page results
    if page:
        click.echo_via_pager(output)
    else:
        ctx.echo(output)


@compute_vmdk_personal.command('sync', short_help='Sync personal VMDK files')
@pass_context
def compute_vmdk_personal_sync(ctx: Configuration):
    """Synchronize VMDK files stored in your personal VSKEY-STOR space.

    Once processed it should be listed with the ls command.
    """
    obj = ctx.sync_user_vmdks()
    columns = ctx.columns or const.COLUMNS_REQUEST_SUBMITTED
    ctx.echo(format_output(ctx, [obj], columns=columns, single=True))



"""Compute Content Library plugin for VSS CLI (vss-cli)."""
import logging

import click

from vss_cli import const, rel_opts as so
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output
from vss_cli.plugins.compute import cli

_LOGGING = logging.getLogger(__name__)


@cli.group('contentlib', short_help='Manage Content Library Items.')
@pass_context
def compute_contentlib(ctx: Configuration):
    """Manage Manage Content Library Items.

    Virtual Machine templates, OVF, ISO and other items.
    """


@compute_contentlib.group('vm', short_help='Browse Virtual Machine Templates')
@pass_context
def compute_contentlib_vm_tmpl(ctx: Configuration):
    """Available ISO images in the VSS central store."""


@compute_contentlib_vm_tmpl.command(
    'ls', short_help='list public VM Templates'
)
@so.filter_opt
@so.sort_opt
@so.all_opt
@so.page_opt
@pass_context
def compute_contentlib_vm_tmpl_ls(
    ctx: Configuration, filter_by, show_all, sort, page
):
    """List available VM templates available in the Content Library.

    Filter by name and sort desc. For example:

    vss-cli compute contentlib vm ls -f name=like,Cent% -s path=asc
    """
    params = dict(sort='name,asc')
    if all(filter_by):
        params['filter'] = ';'.join(filter_by)
    if all(sort):
        params['sort'] = ';'.join(sort)
    # get objects
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_content_library_vm_items(show_all=show_all, **params)
    # format
    columns = ctx.columns or const.COLUMNS_CLIB_ITEMS
    output = format_output(ctx, obj, columns=columns)
    # page results
    if page:
        click.echo_via_pager(output)
    else:
        ctx.echo(output)


@compute_contentlib.group('ovf', short_help='Browse current OVF images')
@pass_context
def compute_contentlib_ovf(ctx: Configuration):
    """Available OVF images available in Content Library."""
    pass


@compute_contentlib_ovf.command('ls', short_help='list public OVF images')
@so.filter_opt
@so.sort_opt
@so.all_opt
@so.page_opt
@pass_context
def compute_contentlib_ovf_ls(
    ctx: Configuration, filter_by, show_all, sort, page
):
    """List available OVF templates available in the Content Library.

    Filter by name and sort desc. For example:

    vss-cli compute contentlib ovf ls -f name=photon -s path=asc
    """
    params = dict(sort='name,asc')
    if all(filter_by):
        params['filter'] = ';'.join(filter_by)
    if all(sort):
        params['sort'] = ';'.join(sort)
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_content_library_ovf_items(show_all=show_all, **params)
    # format
    columns = ctx.columns or const.COLUMNS_CLIB_ITEMS
    output = format_output(ctx, obj, columns=columns)
    # page results
    if page:
        click.echo_via_pager(output)
    else:
        ctx.echo(output)


@compute_contentlib.group('iso', short_help='Browse current ISO images')
@pass_context
def compute_contentlib_iso(ctx: Configuration):
    """Available ISO images available in Content Library."""
    pass


@compute_contentlib_iso.command('ls', short_help='list public ISO images')
@so.filter_opt
@so.sort_opt
@so.all_opt
@so.page_opt
@pass_context
def compute_contentlib_iso_ls(
    ctx: Configuration, filter_by, show_all, sort, page
):
    """List available ISO files available in the Content Library.

    Filter by name and sort desc. For example:

    vss-cli compute contentlib iso ls -f name=ubuntu -s path=asc
    """
    params = dict(sort='name,asc')
    if all(filter_by):
        params['filter'] = ';'.join(filter_by)
    if all(sort):
        params['sort'] = ';'.join(sort)
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_content_library_iso_items(show_all=show_all, **params)
    # format
    columns = ctx.columns or const.COLUMNS_CLIB_ITEMS
    output = format_output(ctx, obj, columns=columns)
    # page results
    if page:
        click.echo_via_pager(output)
    else:
        ctx.echo(output)
